#!/usr/bin/env python3
"""
Program to estimate how many outliers off the y=x axis.
    1. hard cutoff. abs(y-x)<=minDelta
    2. model y-x as a normal distribution, estimate its mean/variance
        then add them up as chi-squared statistic.
    
    If "-i ..." is given, it is regarded as one of the input files (plus the ones in trailing arguments).

Examples:
    %s -i /tmp/Contig315_StKitts_vs_Nevis.tsv --xColumnHeader=StKitts --whichColumnHeader=Nevis
        -s 1.0 -o /tmp/Contig315_StKitts_vs_Nevis.2D.png

"""
import sys, os, math
__doc__ = __doc__%(sys.argv[0])
import numpy, random
from palos import PassingData

def calculateChiSqStatOfDeltaVector(dataVector=None, mean=None, std=None):
    """
    2012.10.14
        adapted from vervet/src/pedigree/DetectWrongLabelByCompKinshipVsIBD.DetectWrongLabelByCompKinshipVsIBD.calculateChiSqStatOfDeltaVector()
    2012.8.22
    """
    import rpy
    medianAbsDelta = None
    noOfNonMissing = None
    chiSqStat = 0
    noOfNonMissing = 0
    chiSqMinusLogPvalue = None
    for i in range(len(dataVector)):
        delta = dataVector[i]
        if delta is not None:
            # Take the absolute value, since it's always P(X>a).
            # A negative z-score will get the wrong portion.
            Z = abs((delta-mean)/std)
            # The log is natural log, not log10.
            # Two-tail, not one-tail.
            logPvalue = rpy.r.pnorm(Z, lower_tail = rpy.r.FALSE, log=rpy.r.TRUE)
            logPvalue += math.log(2)
            
            #pvalue = utils.getZScorePvalue(Z)	#numerical underflow, pvalue=0
            # Fisher's method in combining pvalue of z-score (normalized Z).
            chiSqStat += -2*logPvalue
            noOfNonMissing += 1
    if noOfNonMissing>0:
        # df is 2*noOfNonMissing.
        chiSqMinusLogPvalue = -rpy.r.pchisq(chiSqStat, 2*noOfNonMissing, \
                    lower_tail = rpy.r.FALSE, log=rpy.r.TRUE)
        #chiSqPvalue = stats.chisqprob(chiSqStat, df=2*noOfNonMissing)
        #equivalent to = 1- stats.chi2.cdf(chiSqStat, df) = Prob (x<chiSqStat)
    return PassingData(chiSqStat=chiSqStat, noOfNonMissing=noOfNonMissing, \
                       chiSqMinusLogPvalue=chiSqMinusLogPvalue)

def estimateMeanStdFromData(dataVector=None, excludeTopFraction=0.2):
    """
    2012.10.14
        adapted from vervet/src/pedigree/DetectWrongLabelByCompKinshipVsIBD.DetectWrongLabelByCompKinshipVsIBD.estimateAbsDeltaMeanStd()
    2012.8.22
    """
    sys.stderr.write("Estimating mean&std using the middle %.1f%% of data (n=%s) ..."%\
                    ((1-excludeTopFraction)*100, len(dataVector)))
    noOfRows = len(dataVector)
    import numpy
    # 2012.8.22 draw some histogram to check what data looks like
#		if len(dataVector)>10:
#			outputFname = '%s_kinship_ibd_hist.png'%(self.outputFnamePrefix)
#			yh_matplotlib.drawHist(dataVector, title='', \
#							xlabel_1D="kinship-ibd", xticks=None, \
#							outputFname=outputFname, min_no_of_data_points=10, \
#							needLog=True, \
#							dpi=200, min_no_of_bins=25)
    #dataVector = map(abs, dataVector)	#2012.8.23 no abs
    dataVector.sort()
    startIndex = min(0, int(len(dataVector)*(excludeTopFraction/2))-1)
    stopIndex = int(len(dataVector)*(1-excludeTopFraction/2))
    dataVector = dataVector[startIndex:stopIndex]
    
    data_mean = numpy.mean(dataVector)
    data_std = numpy.std(dataVector)
    
    sys.stderr.write(" mean=%.3f, std=%.3f.\n"%(data_mean, data_std))
    return PassingData(mean=data_mean, std=data_std)


def getZScorePvalue(zscore=None, twoSided=False):
    """
    2012.10.15
        was in pymodule/utils.py
    2012.8.22
        becasue this import wouldn't work. hard to remember:
    
            >>> import scipy
            >>> scipy.stats.norm.sf
            Traceback (most recent call last):
              File "<stdin>", line 1, in <module>
            AttributeError: 'module' object has no attribute 'stats'
        
        zscore could also be a vector (list)
    """
    import scipy.stats as stats
    pvalue = stats.norm.sf(zscore)
    if twoSided:
        pvalue  = pvalue* 2
    return pvalue

def reOrderListOfListByFirstMember(listOfList=None):
    """
    2012.10.29
        this function reorders every list in listOfList in ascending order using values in the 1st list of listOfList.
        And every member of listOfList will be turned into numpy.float array.
    2012.10.25
    """
    firstList = listOfList[0]
    x_ar = numpy.array(firstList, numpy.float)
    #sort x_ar and y_ar must be in the order of x_ar
    indexOfOrderList = numpy.argsort(x_ar)
    returnListOfList = []
    for ls in listOfList:
        ar = numpy.array(ls, numpy.float)
        ar = ar[indexOfOrderList]
        returnListOfList.append(ar)
    return PassingData(listOfList=returnListOfList)

def splineFit(x_ls=None, y_ls=None, no_of_steps=100, needReorderData=True):
    """
    2012.10.25
    http://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.UnivariateSpline.html
    class scipy.interpolate.UnivariateSpline(x, y, w=None, bbox=[None, None], k=3, s=None)[source]
        One-dimensional smoothing spline fit to a given set of data points.
        
        Fits a spline y=s(x) of degree k to the provided x, y data. s specifies the number of knots by specifying a smoothing condition.
        
        Parameters :	
        x : array_like
            1-D array of independent input data. Must be increasing.
        y : array_like
            1-D array of dependent input data, of the same length as x.
        w : array_like, optional
            Weights for spline fitting. Must be positive. If None (default), weights are all equal.
            bbox : array_like, optional
            2-sequence specifying the boundary of the approximation interval. If None (default), bbox=[x[0], x[-1]].
        k : int, optional
            Degree of the smoothing spline. Must be <= 5.
        s : float or None, optional
            Positive smoothing factor used to choose the number of knots. Number of knots will be increased until the smoothing condition is satisfied:
            sum((w[i]*(y[i]-s(x[i])))**2,axis=0) <= s
            If None (default), s=len(w) which should be a good value if 1/w[i] is an estimate of the standard deviation of y[i].
            If 0, spline will interpolate through all data points.

    """
    import scipy.interpolate
    if needReorderData:
        sortData = reOrderListOfListByFirstMember(listOfList=[x_ls, y_ls])
        x_ls = sortData.listOfList[0]
        y_ls = sortData.listOfList[1]
    
    sp = scipy.interpolate.UnivariateSpline(x_ls, y_ls,k=5)
    step = (x_ls[-1]-x_ls[0])/float(no_of_steps)
    n_x_ls = numpy.arange(x_ls[0], x_ls[-1], step)
    n_y_ls = map(sp, n_x_ls)
    return PassingData(x_ls=n_x_ls, y_ls=n_y_ls)

def movingAverage(listOfList=None, no_of_steps=100, needReorderData=True, reduceType=1, minNoOfTotal=10,\
                minValueForFraction=0.8):
    """
    2012.10.29
        a function that calculates moving average of every list in listOfList.
        stepping is applied to the first member in listOfList.
        all member lists are of the same length.
        reduceType
            1: median
            2: mean
            3: fraction that is >=minValueForFraction, only applied to lists after the first one, numpy.median for X
    2012.10.26
    """
    if needReorderData:
        sortData = reOrderListOfListByFirstMember(listOfList=listOfList)
        listOfList = sortData.listOfList
    firstList = sortData.listOfList[0]
    step = (firstList[-1]-firstList[0])/float(no_of_steps)
    stepIndex2Data = {}
    noOfLists = len(listOfList)
    for j in range(len(firstList)):
        x = firstList[j]
        stepIndex = int(x/step)	#figure which bracket/bag all the data from this column should fall into.
        if stepIndex not in stepIndex2Data:
            stepIndex2Data[stepIndex] = PassingData(listOfList=[])
            for i in range(noOfLists):
                stepIndex2Data[stepIndex].listOfList.append([])
        #y = y_ls[j]
        #stepIndex2Data[stepIndex].listOfList[0].append(x)
        for i in range(noOfLists):
            valueAtThatList = listOfList[i][j]
            stepIndex2Data[stepIndex].listOfList[i].append(valueAtThatList)
    
    stepIndexList = sorted(stepIndex2Data)
    
    fractionFunction = lambda ls: sum([a>=minValueForFraction for a in ls])/float(len(ls))
    reduceType2Function = {1: numpy.median, 2: numpy.mean, 3: fractionFunction}
    reduceFunction = reduceType2Function.get(reduceType, numpy.median)
    
    n_x_ls = []
    returnListOfList = []
    for i in range(noOfLists):
        returnListOfList.append([])
    
    for stepIndex in stepIndexList:
        data = stepIndex2Data.get(stepIndex)
        subListOfList = data.listOfList
        if len(subListOfList[0])<minNoOfTotal:
            continue
        
        for i in range(noOfLists):
            if i==0 and reduceType==3:
                _reduceFunction = numpy.median
            else:
                _reduceFunction = reduceFunction
            
            reduce_value = _reduceFunction(subListOfList[i])
            returnListOfList[i].append(reduce_value)
    return PassingData(listOfList=returnListOfList)


def calculate7NumberSummaryForOneList(ls, returnObj=None):
    """
    2012.10.29 moved from variation/src/common.py
    2009-11-25
        calculate a 7-number summary stats for a given list
    """
    from scipy import stats	# for scoreatpercentile/percentileatscore to get quartiles
    if returnObj is None:
        returnObj = PassingData()
    returnObj.minimum = numpy.min(ls)
    returnObj.first_decile = stats.scoreatpercentile(ls, 10)
    returnObj.lower_quartile = stats.scoreatpercentile(ls, 25)
    returnObj.median = stats.scoreatpercentile(ls, 50)
    returnObj.upper_quartile = stats.scoreatpercentile(ls, 75)
    returnObj.last_decile = stats.scoreatpercentile(ls, 90)
    returnObj.maximum = numpy.max(ls)
    return returnObj

class NumberContainer(object):
    """
    #2013.05.24 a structure to scale/normalize any range/series of numbers
        to make them [0,1]
    """
    def __init__(self, minValue=None, maxValue=None):
        self.minValue = minValue
        self.maxValue = maxValue
        self.valueList = []
    
    def addOneValue(self, value=None):
        if value is not None:
            self.valueList.append(value)
            if self.minValue is None or self.minValue > value:
                self.minValue = value
            if self.maxValue is None or self.maxValue < value:
                self.maxValue = value
    
    def normalizeValue(self, value=None):
        """
        2013.05.24
        """
        if self.minValue is not None and self.maxValue is not None:
            return (value-self.minValue)/float(self.maxValue-self.minValue)
        else:
            raise

class DiscreteProbabilityMassContainer(object):
    """
    Examples:
        probabilityMassContainer = DiscreteProbabilityMassContainer(object2proabilityMassDict=self.originalIndividualID2representativeData)
        sampledIndividualID = probabilityMassContainer.sampleObject()
    
    2013.05.26
        function to do sampling
    """
    def __init__(self, object2proabilityMassDict=None):
        """
        2013.05.26
        """
        from palos.algorithm.RBTree import RBDict
        self.rbDict = RBDict()
        self.totalProbabilityMass = 1	#default
        if object2proabilityMassDict is not None:
            self._constructFromDiscreteProbabilityMassDict(dc=object2proabilityMassDict)
    
    def _constructFromDiscreteProbabilityMassDict(self, dc=None):
        """
        2013.05.28
            dc is a structure with object name as key, and object probability mass (normalized or not) as value. i.e.
                {"1978001":0.5, "1980001":1.5}
                
            argument probabilityNormalized: whether the sum of all values in dc adds up to 1.
        """
        from palos.polymorphism.CNV import CNVSegmentBinarySearchTreeKey
        startProbMass = 0.0
        for discreteVariable, probabilityMass in dc.items():
            segmentKey = CNVSegmentBinarySearchTreeKey(chromosome="1", span_ls=[startProbMass, startProbMass+probabilityMass], \
                                                min_reciprocal_overlap=0.001, isDataDiscrete=False)
                        #min_reciprocal_overlap=1: must be complete overlap in order for two objects occupying same key
            self.rbDict[segmentKey] = discreteVariable
            startProbMass += probabilityMass
        self.totalProbabilityMass = startProbMass
        sys.stderr.write("%s\n"%(repr(self.rbDict)))
    
    def sampleObject(self):
        """
        """
        from palos.polymorphism.CNV import CNVSegmentBinarySearchTreeKey
        u = random.random()*self.totalProbabilityMass
        key = CNVSegmentBinarySearchTreeKey(chromosome="1", span_ls=[u], \
                                                min_reciprocal_overlap=0.0000001)
        #randint.(0,noOfTotalRows-1)
        
        node = self.rbDict.findNode(key)
        if node:
            return node.value
        else:
            return None
    
        


#!/usr/bin/env python3
"""
abstract mapper for Sunset mappers/reducers.
"""
from . AbstractDBJob import AbstractDBJob
import sys, os
import copy
import logging
from palos import ProcessOptions
from palos.db import SunsetDB

class AbstractSunsetMapper(AbstractDBJob):
    __doc__ = __doc__
    option_default_dict = copy.deepcopy(AbstractDBJob.option_default_dict)
    #option_default_dict.pop(('inputFname', 0, ))
    
    def __init__(self, inputFnameLs=None, **keywords):
        """
        """
        #self.connectDB() called within __init__()
        AbstractDBJob.__init__(self, inputFnameLs=inputFnameLs, **keywords)
    
    def connectDB(self):
        """
        """
        self.db_main = SunsetDB.SunsetDB(drivername=self.drivername, 
            hostname=self.hostname,
            dbname=self.dbname, schema=self.schema, port=self.port,
            db_user=self.db_user, db_passwd=self.db_passwd)
        self.db_main.setup(create_tables=False)
    
    def checkIfAlignmentListMatchMethodDBEntry(self, 
        individualAlignmentLs:list=None, methodDBEntry=None, session=None):
        """
        moved from AddVCFFile2DB.py
        """
        #make sure methodDBEntry.individual_alignment_ls is identical to 
        # individualAlignmentLs
        alignmentIDSetInFile = set([alignment.id for alignment in individualAlignmentLs])
        alignmentIDSetInGenotypeMethod = set([alignment.id \
            for alignment in methodDBEntry.individual_alignment_ls])
        if alignmentIDSetInFile!=alignmentIDSetInGenotypeMethod:
            logging.error(f"alignmentIDSetInFile ({repr(alignmentIDSetInFile)}) "
                f"doesn't match alignmentIDSetInFile "
                f"({repr(alignmentIDSetInGenotypeMethod)}).")
            if session:
                session.rollback()
            #delete all target files if there is any
            self.cleanUpAndExitOnFailure(exitCode=2)
            
if __name__ == '__main__':
    main_class = AbstractSunsetMapper
    po = ProcessOptions(sys.argv, main_class.option_default_dict,
        error_doc=main_class.__doc__)
    instance = main_class(**po.long_option2value)
    instance.run()



#!/usr/bin/env python3
"""
Description:
    Abstract class for programs that walk through a list of matrix-like files.
    Running it will reformat input into tab-delimited tsv matrix file.
    If whichColumn/whichColumnHeader is given, it'll convert it into float or log or -log.
    It combines both input from "-i oneFile.txt" or the trailing standalone arguments.
    It could act as both a mapper (one input) or reducer (multiple input).

"""

import sys, os, math
import random
import logging
from palos import ProcessOptions, getListOutOfStr, PassingData, utils
from palos.mapper.AbstractMapper import AbstractMapper
from palos.io.MatrixFile import MatrixFile
from palos.io.HDF5MatrixFile import HDF5MatrixFile
from palos.io.YHPyTables import YHFile
from palos.ngs.io.VCFFile import VCFFile

class AbstractMatrixFileWalker(AbstractMapper):
    __doc__ = __doc__
    option_default_dict = AbstractMapper.option_default_dict.copy()
    #option_default_dict.update(AbstractMapper.db_option_dict.copy())
    option_default_dict.update({
        ('minNoOfTotal', 1, int): [10, 'M', 1, 'minimum no of data from one file for afterFileFunction() to run'],\
        ('h5TableName', 0, ): [None, '', 1, 'table name in the input file if inputFileFormat is 2'],\
        ('maxNoOfTotal', 0, int): [None, '', 1, 'maximum no of data to sample from one file. if not set, no limit.'],\
        ('samplingRate', 1, float): [1, 's', 1, 'how often you include the data, a probability between 0 and 1.'],\
        ('whichColumn', 0, int): [None, 'w', 1,
            'data from this column (index starting from 0) is plotted as y-axis value'],\
        ('whichColumnHeader', 0, ): [None, 'W', 1, 
            'column header for the data to be plotted as y-axis value, substitute whichColumn'],\
        ('logY', 0, int): [0, '', 1,
            'value 0: nothing; 1: log(); 2: -log(); 3: ln(X); 4: -ln(X). replacing self.logWhichColumn.'],\
        ('valueForNonPositiveYValue', 1, float): [-1, 'v', 1, 
            'default value when log-transformation fails (when value is negative)'],\
        ('missingDataNotation', 0, ): ['NA', '', 1, 
            'coma-separated list of notations for missing data.'
            'missing data will be skipped.'],\
        ('inputFileFormat', 0, int): [1, '', 1, '1: csv-like plain text file; '
            '2: YHPyTables.YHFile; 3: HDF5MatrixFile; 4: VCFFile'],\
        ('outputFileFormat', 0, int): [1, '', 1, '1: csv-like plain text file; '
            '2: YHPyTables.YHFile; 3: HDF5MatrixFile; 4: csv-like matrix, without header'],\
        })
    
    def __init__(self, inputFnameLs=None, **keywords):
        """
        self.missingDataNotation will be processed into a set.
        """
        #self.connectDB() called within its __init__()
        AbstractMapper.__init__(self, inputFnameLs=inputFnameLs, **keywords)
        #if user wants to preserve data in a data structure that is visible throughout reading different files.
        # then use this self.invariantPData.
        self.invariantPData = PassingData(writer=None, headerOutputted=False,
            x_ls = [], y_ls = [], z_ls=[])
        if getattr(self, 'missingDataNotation', None):
            self.missingDataNotation = set(utils.getListOutOfStr(
                self.missingDataNotation, data_type=str, separator2=None))
        else:
            self.missingDataNotation = set()
    
    def connectDB(self):
        """
        split out of __init__() so that derived classes could overwrite this function
        """
        pass
    
    def preFileFunction(self, **keywords):
        """
        """
        if hasattr(self, 'initiatePassingData'):
            return self.initiatePassingData(**keywords)
    
    def initiatePassingData(self, ):
        """
        this function gets called in the beginning of each fileWalker() (for each inputFname)
        """
        pdata = PassingData(x_ls = [], y_ls = [], invariantPData=self.invariantPData)
        #2012.8.16 pass to global data
        self.invariantPData.y_ls = pdata.y_ls
        self.invariantPData.x_ls = pdata.x_ls
        return pdata
    
    def processValue(self, value=None, processType=None,
        valueForNonPositiveValue=None, **keywords):
        """
        processType 0: nothing; 1: log(), 2: -log().
        2012.10.15
            the default value of takeLogarithm depends on (self.logWhichColumn or self.logY)
        """
        if valueForNonPositiveValue is None:
            valueForNonPositiveValue = self.valueForNonPositiveYValue
        value = float(value)
        if processType is not None and processType>=1:
            if value>0:
                if processType==1:
                    value = math.log10(value)
                elif processType==2:
                    value = -math.log10(value)
            else:
                value = valueForNonPositiveValue
        return value
    
    def handleYValue(self, yValue=None, processType=None,
        valueForNonPositiveValue=None, **keywords):
        """
        """
        return self.processValue(value=yValue, processType=processType,
            valueForNonPositiveValue=valueForNonPositiveValue, **keywords)
    
    def _writeHeader(self, header=None, pdata=None, rowDefinition=None):
        """
        called by processHeader() and others (in GenomeMovingAverageStatistics.py)
        """
        if not self.invariantPData.headerOutputted:
            if self.outputFileFormat==1:
                if self.invariantPData.writer and header:
                    self.invariantPData.writer.writerow(header)
            elif getattr(self, 'writer', None) is None and \
                getattr(self.invariantPData, 'writer', None) is None:
                if self.outputFileFormat==2:
                    if not rowDefinition and header:
                        #generate a rowDefinition based on header
                        rowDefinition = []
                        for colID in header:
                            rowDefinition.append((colID, 's2000'))
                    writer = YHFile(self.outputFname, mode='w', rowDefinition=rowDefinition)
                    self.invariantPData.writer = writer
                else:	#for HDF5MatrixFile
                    if not rowDefinition and header:
                        #generate a rowDefinition based on header
                        rowDefinition = []
                        for colID in header:
                            rowDefinition.append((colID, HDF5MatrixFile.varLenStrType))
                    #rowDefinition = [('locus_id','i8'),
                    #   ('chromosome', HDF5MatrixFile.varLenStrType), ('start','i8'), ('stop', 'i8'),
                    #	('score', 'f8'), ('MAC', 'i8'), ('MAF', 'f8')]
                    writer = HDF5MatrixFile(self.outputFname, mode='w', rowDefinition=rowDefinition)
                    self.invariantPData.writer = writer
            else:
                logging.warn("Either self.writer %s, or self.invariantPData.writer %s already exists."%\
                    (getattr(self, 'writer', None), getattr(self.invariantPData, 'writer', None)))
                logging.warn("\t no writer created in processHeader().")
        self.invariantPData.headerOutputted = True
    
    def processHeader(self, header=None, pdata=None, rowDefinition=None):
        """
        only open self.outputFname for write when self.writer and self.invariantPData.writer is not available
        called right after the header of an input file is derived in fileWalker().
        """
        self._writeHeader(header=header, pdata=pdata, rowDefinition=rowDefinition)
    
    
    def processRow(self, row=None, pdata=None):
        """
        return returnValue: 0 if not included or nothing is done on it.
            1 if included or something is carried out on it.
        handles each row in each file, here it replaces the yValue.
        """
        returnValue = 0
        col_name2index = getattr(pdata, 'col_name2index', None)
        y_ls = getattr(pdata, 'y_ls', None)
        if col_name2index and y_ls is not None:
            if self.whichColumnHeader:
                whichColumn = col_name2index.get(self.whichColumnHeader, None)
            elif self.whichColumn:
                whichColumn = self.whichColumn
            else:
                whichColumn = None
            if whichColumn is not None:
                yValue = row[whichColumn]
                if yValue not in self.missingDataNotation:
                    yValue = self.processValue(yValue, processType=self.logY,
                        valueForNonPositiveValue=self.valueForNonPositiveYValue)
                row[whichColumn] = yValue
            if self.invariantPData.writer:
                self.invariantPData.writer.writerow(row)
                returnValue = 1
        return returnValue
    
    def getNumberOfData(self, pdata):
        """
        """
        return len(pdata.y_ls)
        
    def afterFileFunction(self, **keywords):
        """
        """
        if hasattr(self, 'plot'):
            return self.plot(**keywords)
    
    def openOneInputFile(self, inputFname=None):
        """
        2013.09.05 split out of fileWalker() , added VCFFile
        """
        if self.inputFileFormat==2:
            reader = YHFile(inputFname, mode='r', tableName=self.h5TableName)
        elif self.inputFileFormat==3:
            reader = HDF5MatrixFile(inputFname, mode='r')
        elif self.inputFileFormat==4:
            reader = VCFFile(inputFname=inputFname)
        else:
            reader = MatrixFile(inputFname)
        return reader
    
    def fileWalker(self, inputFname=None, preFileFunction=None,
        afterFileFunction=None, processRowFunction=None , run_type=1):
        """
        inputFileFormat==2, support HDF5MatrixFile
        """
        sys.stderr.write("walking through %s ..."%(inputFname))
        counter = 0
        real_counter = 0
        noOfSampled = 0
        if preFileFunction is None:
            preFileFunction = self.preFileFunction
        pdata = preFileFunction()
        if processRowFunction is None:
            processRowFunction = self.processRow
        if afterFileFunction is None:
            afterFileFunction = self.afterFileFunction
        try:
            reader = self.openOneInputFile(inputFname)
            col_name2index = reader.constructColName2IndexFromHeader()
            header = reader.getHeader()
            
            #output the header to the output file if necessary 
            self.processHeader(header=header, pdata=pdata)
            pdata.reader = reader
            pdata.col_name2index = col_name2index
            pdata.filename = inputFname
            
            for row in reader:
                counter += 1
                if self.samplingRate<1 and self.samplingRate>=0:
                    r = random.random()
                    if r>self.samplingRate:
                        continue
                rowReturnValue = processRowFunction(row=row, pdata=pdata)
                if rowReturnValue is None:
                    rowReturnValue = 0
                real_counter += rowReturnValue
                noOfSampled += 1
                if self.maxNoOfTotal and real_counter>self.maxNoOfTotal:
                    break
            if self.getNumberOfData(pdata)>=self.minNoOfTotal:
                afterFileFunction(x_ls=pdata.x_ls, y_ls=pdata.y_ls, pdata=pdata)
            del reader
        except:
            sys.stderr.write('Except type: %s\n'%repr(sys.exc_info()))
            import traceback
            traceback.print_exc()
            sys.exit(3)
        if counter>0:
            fraction = float(noOfSampled)/float(counter)
        else:
            fraction = 0
        sys.stderr.write("%s/%s (%.3f) data sampled. real_counter=%s.\n"%(
            noOfSampled, counter, fraction, real_counter))
    
    def setup(self, **keywords):
        """
        do not open the file if it's a png file.
        run before anything is run.
        """
        writer = None
        if self.outputFileFormat in [1,4]:
            suffix = os.path.splitext(self.outputFname)[1]
            if self.outputFname and suffix!='.png':
                writer = MatrixFile(self.outputFname, mode='w', delimiter='\t')
        else:
            #HDF5MatrixFile
            #can't generate HDF5MatrixFile, because it needs dtypeList
            pass
        #pass it to the invariantPData
        self.invariantPData.writer = writer
        self.writer = writer
    
    def closeFiles(self, **keywords):
        """
        2013.1.27
        """
        if getattr(self.invariantPData, 'writer', None):
            try:
                self.invariantPData.writer.close()
            except:
                sys.stderr.write('Except type: %s\n'%repr(sys.exc_info()))
                import traceback
                traceback.print_exc()
            
            #this would close file too if it's not closed
            del self.invariantPData.writer
        
        if getattr(self, 'writer', None):
            del self.writer
        
    def reduce(self, **keywords):
        """
        run after all files have been walked through
        """
        pass
        
    
    def run(self):
        
        if self.debug:
            import pdb
            pdb.set_trace()
        
        self.setup()
        
        for inputFname in self.inputFnameLs:
            if os.path.isfile(inputFname):
                self.fileWalker(inputFname, afterFileFunction=self.afterFileFunction,
                    run_type=1, processRowFunction=self.processRow)
        
        self.reduce()
        self.closeFiles()
        
            
if __name__ == '__main__':
    main_class = AbstractMatrixFileWalker
    po = ProcessOptions(sys.argv, main_class.option_default_dict, error_doc=main_class.__doc__)
    instance = main_class(po.arguments, **po.long_option2value)
    instance.run()


#!/usr/bin/env python3g

import os
import sys
from setuptools import setup, find_packages

src_dir = os.path.dirname(__file__)

install_requires = ['h5py', 'matplotlib', "numpy", 
    'scipy', "sqlalchemy", 'statsmodels', 'tables',
    'future', 'future-fstrings']

if sys.version_info.major>2:
    #biopython, Pegaflow is Python3 only
    install_requires.append("biopython")
    install_requires.append("pegaflow")

with open(os.path.join(src_dir, 'README.md')) as readme_file:
    README = readme_file.read()
#
# Create Manifest file to exclude tests, and service files
#
def create_manifest_file():
    f = None
    try:
        f = open('MANIFEST.in', 'w')
        f.write(u"global-exclude *.py[cod]\n")
    finally:
        if f:
            f.close()

#
# Install conditional dependencies
#
def setup_installer_dependencies():
    global install_requires

def find_package_data(dirname):
    def find_paths(dirname):
        items = []
        for fname in os.listdir(dirname):
            path = os.path.join(dirname, fname)
            if os.path.isdir(path):
                items += find_paths(path)
            elif not path.endswith(".py") and not path.endswith(".pyc"):
                items.append(path)
        return items

    items = find_paths(dirname)
    return [path.replace(dirname, "") for path in items]




if __name__ == '__main__':
    create_manifest_file()
    setup_installer_dependencies()
    setup(
        name="Palos",
        version="0.1.35",
        author="Yu S. Huang",
        author_email="polyactis@gmail.com",
        description="Misc Python modules developed and used by the yfish group",
        long_description_content_type="text/markdown",
        long_description=README,
        license="Apache2",
        url="https://github.com/polyactis/pymodule",
        python_requires=">=2.7",
        classifiers=[
            "Intended Audience :: Developers",
            "Intended Audience :: Science/Research",
            "Operating System :: Unix",
            "Programming Language :: Python",
            "Programming Language :: Python :: 3",
            "Topic :: Scientific/Engineering",
            "Topic :: Utilities",
        ],
        packages=find_packages(include=['palos', 'palos.*'], exclude=['.test*']),
        package_data={
            # If any package contains *.sh files, include them:
            "": ["*.sh", "*.md"],
        },
        include_package_data=True,
        zip_safe=False,
        install_requires = install_requires,
    )



"""
2013.3.6 data structure for polymorphism data in YHPyTables (custom of PyTables) format
	testing examples could be found in ~/script/pymodule/popgen/examples
"""
import sys, os, math
import csv
import tables
import numpy
import networkx as nx
from tables import UInt64Col, Float64Col, StringCol, UInt16Col, Int64Col
from palos.utils import PassingData, PassingDataList
from palos.ProcessOptions import  ProcessOptions
from palos.io.YHPyTables import YHTable, YHFile, castPyTablesEntryIntoPassingData
from palos.polymorphism.SNP import SNPData

class SpeciesTable(tables.IsDescription):
	"""
	2013.3.8
	"""
	id = UInt64Col(pos=0)
	name = StringCol(512, pos=1)
	scientific_name = StringCol(512, pos=2)
	ploidy = UInt16Col(pos=3)

class PopulationTable(tables.IsDescription):
	"""
	2013.3.8
	"""
	id = UInt64Col(pos=0)
	name = StringCol(512, pos=1)	#name should like 'species_name.population_name'
	size = UInt64Col(pos=2)
	species_id = UInt64Col(pos=3)

class ChromosomeTable(tables.IsDescription):
	"""
	2013.3.8
	"""
	id = UInt64Col(pos=0)
	name = StringCol(512, pos=1)	#should be unique, but not 100% enforced
		#name should look like species_name.chromosome_name, if species_name is available
	length = UInt64Col(pos=2)
	sex_chromosome = StringCol(4, pos=3)	#A=autosome, X=chr X, Y=chr Y
	species_id = UInt64Col(pos=4)
	path = StringCol(4512, pos=5)	#path to the file containing the chromosome sequences, should be unique, 

class LocusTable(tables.IsDescription):
	id = UInt64Col(pos=0)
	name = StringCol(512, pos=1)	#'species_name.chromosome_name.locus_name'
	chromosome_id = UInt64Col(pos=2)	#64 byte-long
	start = UInt64Col(pos=3)
	stop = UInt64Col(pos=4)
	ref_allele = StringCol(64, pos=5)
	ref_allele_length = UInt64Col(pos=6)	#this could be larger than 64, good when ref_allele is simply too long
	ref_allele_frequency = Float64Col(pos=7)
	
	alt_allele = StringCol(64, pos=8)
	alt_allele_length = UInt64Col(pos=9)	#this could be larger than 64, good when ref_allele is simply too long
	alt_allele_frequency = Float64Col(pos=10)
	
	generation_mutation_arose = Int64Col(pos=11)
	generation_mutation_fixed = Int64Col(pos=12)
	mutation_type = StringCol(4, pos=13)	#0=synonymous/non-coding, 1=non-synonymous, i=insertion, d=deletion, v=inversion
	fitness = Float64Col(pos=14)
	ancestral_amino_acid = StringCol(4, pos=15)	#only for mutation_type 0/1
	derived_amino_acid = StringCol(4, pos=16)
	

class IndividualTable(tables.IsDescription):
	id = UInt64Col(pos=0)
	family_id = StringCol(512, pos=1)	#64 byte-long
	name = StringCol(512, pos=2)	# name should look like 'species_name.population_name.individual_name' to ensure uniqueness
	father_name = StringCol(512, pos=3)
	mother_name = StringCol(512, pos=4)
	sex = UInt64Col(pos=5)	#0 is unknown, 1=male, 2=female
	phenotype = Float64Col(pos=6)
	population_id = UInt64Col(pos=7)
	

class PolymorphismTable(tables.IsDescription):
	"""
	2013.3.6 table that records the polymorphism alleles of individuals at relevant loci
	"""
	id = UInt64Col(pos=0)
	name = StringCol(512, pos=1)	#name should be "individualName.locusName.chromosome_copy" to ensure uniqueness
	individual_id = UInt64Col(pos=2)
	locus_id = UInt64Col(pos=3)
	chromosome_copy = UInt64Col(pos=4, dflt=0)	#starting from 0
	allele_sequence = StringCol(512, pos=5)
	allele_sequence_length = UInt64Col(pos=6)
	allele_type = StringCol(512, pos=7)

class OneIndividualPolymorphismData(PassingData):
	"""
	2013.3.8 to store an individual's polymorphism data, i.e. haplotypes or genotypes
		used in SimulatePedigreeHaplotype.py
	"""
	def __init__(self, **keywords):
		self.isPhased = None
		self.ploidy = None
		self.locusIDList = []
		self.haplotypeList = []
		self.locusPositionList = []
		PassingData.__init__(self, **keywords)
	
	def addHaplotype(self, haplotype=None):
		self.haplotypeList.append(haplotype)


class RecombinationTable(tables.IsDescription):
	"""
	2013.3.7 in case a pedigree is stored in IndividualTable, this table records the recombination events in meiosis.
		Used in SimulatePedigreeHaplotype.py
	"""
	id = UInt64Col(pos=0)
	parent_id = UInt64Col(pos=1)
	child_id = UInt64Col(pos=2)
	position = UInt64Col(pos=3, dflt=0)

class PolymorphismTableFile(YHFile):
	"""
	2013.3.6 usage examples:
	#for writing
	landscapeFile = AssociationLandscapePyTable(self.outputFname, mode='w', isPhased=1, ploidy=2)
	landscapeFile.addAttributeDict(attributeDict)
	
	#for read-only,
	landscapeTable = AssociationLandscapePyTable(self.path, mode='r')
	"""
	def __init__(self, path=None, mode='r', \
		tableName='polymorphism', groupNamePrefix='group', tableNamePrefix='table',\
		filters=None, autoRead=True, autoWrite=True, \
		isPhased=None, ploidy=None, constructSNPData=True, **keywords):
		
		
		self.bridge_ls = None
		self.locusLandscapeNeighborGraph = None
		
		YHFile.__init__(self, path=path, mode=mode, \
				tableName=tableName, groupNamePrefix=groupNamePrefix, tableNamePrefix=tableNamePrefix,\
				rowDefinition=None, filters=filters, \
				debug=0, report=0, autoRead=False, autoWrite=False)
		
		self.speciesTableName = 'species'
		self.populationTableName = 'population'
		self.individualTableName = "individual"
		self.chromosomeTableName = 'chromosome'
		self.locusTableName = 'locus'
		self.recombinationTableName = 'recombination'
		
		self.isPhased = isPhased
		self.ploidy = ploidy
		self.constructSNPData = constructSNPData
		
		#to overwrite self.autoRead that is set by YHFile.__init__
		self.autoRead = autoRead
		self.autoWrite = autoWrite
		
		self.snpData = None	#the SNPData structure that holds all polymorphism, locus, individual info
		
		if self.autoRead and (self.mode=='r' or self.mode=='a'):
			self.speciesTable = self.getTableObject(tableName=self.speciesTableName)
			self.populationTable = self.getTableObject(tableName=self.populationTableName)
			self.individualTable = self.getTableObject(tableName=self.individualTableName)
			self.chromosomeTable = self.getTableObject(tableName=self.chromosomeTableName)
			self.locusTable = self.getTableObject(tableName=self.locusTableName)
			self.recombinationTable = self.getTableObject(tableName=self.recombinationTableName)
			self.polymorphismTable = self.getTableObject(tableName=self.tableName)
			
			#read the isPhased, ploidy from pytables attributes, overwrites the arguments
			self.isPhased = self.polymorphismTable.getAttribute(name='isPhased', defaultValue=0)
			self.ploidy = self.polymorphismTable.getAttribute(name='ploidy', defaultValue=2)
			
			self._readInData(tableName=self.tableName, tableObject=self.associationLandscapeTable)
		if self.autoWrite and self.mode=='w':
			self.speciesTable = self.createNewTable(tableName=self.speciesTableName, rowDefinition=SpeciesTable,\
													expectedrows=500)
			self.populationTable = self.createNewTable(tableName=self.populationTableName, rowDefinition=PopulationTable,\
													expectedrows=500)
			self.individualTable = self.createNewTable(tableName=self.individualTableName, rowDefinition=IndividualTable,\
													expectedrows=30000)
			self.chromosomeTable = self.createNewTable(tableName=self.chromosomeTableName, rowDefinition=ChromosomeTable,\
													expectedrows=500)
			self.locusTable = self.createNewTable(tableName=self.locusTableName, rowDefinition=LocusTable,\
												expectedrows=300000)
			self.recombinationTable = self.createNewTable(tableName=self.recombinationTableName, rowDefinition=RecombinationTable,\
												expectedrows=300000)
			self.polymorphismTable = self.createNewTable(tableName=self.tableName, rowDefinition=PolymorphismTable,\
												expectedrows=500000)
			#set the attributes of isPhased, ploidy
			self.polymorphismTable.addAttribute(name='isPhased', value=self.isPhased, overwrite=True)
			self.polymorphismTable.addAttribute(name='ploidy', value=self.ploidy, overwrite=True)
		
		#2013.3.8 these dictionaries are for outputting purposes 
		self._individualName2ID = {}
		self._locus_index2id = {}
		
		#2013.3.8 helper structures
		self._locusStartPositionList = None
		self._locusChrStartStopList = None
	
	def _readInData(self, tableName=None, tableObject=None, bugfixType=None):
		"""
		2013.3.6
		"""
		if tableName is None:
			tableName = self.tableName
		YHFile._readInData(self, tableName=tableName, tableObject=tableObject)
		if not self.constructSNPData:
			return
		
		sys.stderr.write("Reading everything into a SNPData structure ...")
		row_id_list = []
		row_id_number2row_index = {}
		col_id_list = []
		col_id_number2col_index = {}
		for row in self.individualTable:
			row_id_list.append(row.name)
			row_id_number2row_index[row.id] = len(row_id_list)-1
		for row in self.locusTable:
			#col_id_list.append(row.id)
			col_id_list.append((row.chromosome_id, row.start, row.stop))
			col_id_number2col_index[row.id] = len(col_id_list)-1
		
		allele_sequence2allele_number = {}
		allele_number2allele_sequence = {}
		
		#each cell in data_matrix is an array of alleles for one individual at one locus, but different chromosomes
		# alleles are encoded in numbers starting from 1. 0 is missing.
		data_matrix = numpy.zeros([len(row_id_list), len(col_id_list), self.ploidy], dtype=numpy.int16)
		
		if self.ploidy>1:
			#chromosome_copy_matrix is used to keep track of the chromosomes for particular individual & locus
			chromosome_copy_matrix = numpy.zeros([len(row_id_list), len(col_id_list)], dtype=numpy.int8)
		else:
			chromosome_copy_matrix = None
		
		for row in self.polymorphismTable:
			row_index = row_id_number2row_index.get(row.individual_id)
			col_index = col_id_number2col_index.get(row.locus_id)
			
			#figure out which chromosome to hold this allele
			if self.ploidy>1:
				chromosome_copy_matrix[row_index][col_index] = chromosome_copy_matrix[row_index][col_index]+1
				if row.chromosome_copy == 0:	#unphased genotype
					chromosome_copy_index = chromosome_copy_matrix[row_index][col_index] -1
				else:
					chromosome_copy_index = row.chromosome_copy-1
			else:
				chromosome_copy_index = 0
				if row.chromosome_copy>1:
					sys.stderr.write("Warning: ploidy=%s, but encounter chromosome_copy (%s) >1.\n"%\
									(self.ploidy, row.chromosome_copy))
			
			#allele_number starts from 1. 0 is reserved for missing.
			if row.allele_sequence not in allele_sequence2allele_number:
				allele_sequence2allele_number[row.allele_sequence] = len(allele_sequence2allele_number)+1
				allele_number = allele_sequence2allele_number.get(row.allele_sequence)
				allele_number2allele_sequence[allele_number] = row.allele_sequence
			
			allele_number = allele_sequence2allele_number.get(row.allele_sequence)
			data_matrix[row_index][col_index][chromosome_copy_index] = allele_number
		self.snpData = SNPData(row_id_list=row_id_list, col_id_list=col_id_list, data_matrix=data_matrix)
		
		self.snpData.allele_sequence2allele_number = allele_sequence2allele_number
		self.snpData.allele_number2allele_sequence = allele_number2allele_sequence
		sys.stderr.write(" %s individuals, %s loci, ploidy=%s, isPhased=%s.\n"%(len(self.snpData.row_id_ls),\
																	len(self.snpData.col_id_ls), \
																	self.ploidy, self.isPhased))
		
		return self.snpData
	
	def getOneHaplotype(self, individual_id=None, chromosome_copy=None):
		"""
		2013.3.7
			if data is unphased and multi-ploidy, the allele combination is entirely random.
			if data is phased, and multi-ploidy, it will randomly select one chromosome copy in advance.
		"""
		if chromosome_copy is None and self.isPhased:	#randomly select one copy if data is phased
			if self.ploidy>=2:	#
				chromosome_copy  = numpy.random.randint(1, self.ploidy+1, size=1)	#sample one integer from [1, self.ploidy]
				
			else:
				chromosome_copy = None	#do not add it to query restriction condition
		where_condition_ls = ["individual_id==%s"%(individual_id)]
		if chromosome_copy is not None:
			where_condition_ls.append(("chromosome_copy==%s"%(chromosome_copy)))
		query = self.polymorphismTable.query(""" %s """%(" & ".join(where_condition_ls)))
		haplotype = ['']*self.locusTable.nrows
		for row in query:
			locus_index =row['locus_id']-1
			haplotype[locus_index] = row['allele_sequence']
		return haplotype

	def getOneIndividualGenotypeList(self, individual_id=None):
		"""
		2013.3.7
			to get unphased genotype data
		"""
		where_condition_ls = ["individual_id==%s"%(individual_id)]
		query = self.polymorphismTable.query(""" %s """%(" & ".join(where_condition_ls)))
		genotypeList = ['']*self.locusTable.nrows
		for row in query:
			locus_index =row['locus_id']-1
			genotypeList[locus_index] = genotypeList[locus_index] + row['allele_sequence']
		return genotypeList
		
	def sampleOneRandomHaplotypeWithReplacement(self):
		"""
		2013.3.7 randomly select a haplotype and return
		"""
		no_of_individuals = self.individualTable.no_of_rows
		random_individual_id = numpy.random.randint(1, no_of_individuals+1, size=1)
		return self.getOneHaplotype(individual_id=random_individual_id, chromosome_copy=None)
	
	def sampleOneIndividualPolymorphismWithReplacement(self):
		"""
		2013.3.7 randomly select a haplotype and return
		"""
		no_of_individuals = self.individualTable.no_of_rows
		random_individual_id = numpy.random.randint(1, no_of_individuals+1, size=1)
		polymorphismData = OneIndividualPolymorphismData(isPhased=self.isPhased, ploid=self.ploidy)
		if self.isPhased and self.ploidy>0:
			for i in range(1,self.ploidy+1):
				haplotype = self.getOneHaplotype(individual_id=random_individual_id, chromosome_copy=i)
				polymorphismData.addHaplotype(haplotype)
		else:
			genotypeList = self.getOneIndividualGenotypeList(individual_id=random_individual_id)
			polymorphismData.addHaplotype(genotypeList)
		return polymorphismData
	
	@property
	def locusStartPositionList(self):
		"""
		2013.3.7 return a list of locus.start
		"""
		if self._locusStartPositionList is None:
			self._locusStartPositionList = []
			for row in self.locusTable:
				self._locusStartPositionList.append(row['start'])
			self._locusStartPositionList.sort()
		return self._locusStartPositionList
	
	@property
	def locusChrStartStopList(self):
		"""
		2013.3.8 return a list of (locus.chromosome, locus.start, locus.stop)
		"""
		if self._locusChrStartStopList is None:
			self._locusChrStartStopList = []
			for row in self.locusTable:
				chromosomeEntry = self.getChromosome(id=row['chromosome_id'])
				self._locusChrStartStopList.append((chromosomeEntry.name, row['start'], row['stop']))
			self._locusChrStartStopList.sort()
		return self._locusChrStartStopList
	
	def checkIfEntryInTable(self, tableObject=None, name=None, entry_id=None):
		"""
		2013.3.8
			for Population, Species, Chromosome, similar structure
		"""
		returnValue = None
		if tableObject and (name or entry_id):
			where_condition_ls = []
			if name:
				where_condition_ls.append("name=='%s'"%(name))
			if entry_id:
				where_condition_ls.append("id==%s"%(entry_id))
			rows = tableObject.readWhere("""%s"""%(' & '.join(where_condition_ls)))
			if len(rows)==1:
				returnValue = rows[0]
			elif len(rows)>1:
				sys.stderr.write("Error: %s (>1) %s entries with same name.\n"%\
					(len(rows), tableObject.name))
				raise
			else:
				returnValue = False
		if returnValue:
			returnValue = castPyTablesEntryIntoPassingData(returnValue)
		return returnValue
	
	def checkChromosome(self, name=None, id=None):
		"""
		2013.3.8
		"""
		return self.checkIfEntryInTable(tableObject=self.chromosomeTable, \
			name=name, entry_id=id)
	
	def addChromosome(self, name=None, length=None, speciesName=None, ploidy=None, path=None):
		"""
		2013.3.8
		"""
		if name:
			species_id = None
			if speciesName:
				speciesEntry = self.getSpecies(name=speciesName, ploidy=ploidy)
			if speciesEntry:
				species_id = speciesEntry.id
			oneCell = PassingData(name=name, length=length, species_id=species_id, path=path)
			self.chromosomeTable.writeOneCell(oneCell, cellType=2)
			self.flush()
		return self.checkChromosome(name=name)	#would this work without flush()?
	
	def getChromosome(self, name=None, id=None, length=None, speciesName=None, path=None, \
					ploidy=None, getRandomChromosome=False):
		"""
		2013.3.8
			#. if name is None, try get 1st entry in chromosome table
			#. ploidy is for species
		"""
		entry = None
		if name or id:
			entry= self.checkChromosome(name=name, id=id)
			if not entry:
				entry = self.addChromosome(name=name, length=length, speciesName=speciesName, ploidy=ploidy, path=path)
		elif getRandomChromosome:
			entry = self.getFirstChromosome()
		return entry
	
	def getFirstChromosome(self):
		"""
		2013.3.8
		"""
		entry = None
		rows = self.chromosomeTable.read(start=0, stop=1)
		if len(rows)==1:
			entry= rows[0]
		return entry
	
	def checkPopulation(self, name=None, id=None):
		"""
		2013.3.8
		"""
		return self.checkIfEntryInTable(tableObject=self.populationTable, 
			name=name, entry_id=id)
	
	def addPopulation(self, name=None, size=None, speciesName=None):
		"""
		2013.3.8
		"""
		if name:
			species_id = None
			if speciesName:
				species = self.getSpecies(name=speciesName)
			if species:
				species_id = species.id
			oneCell = PassingData(name=name, size=size, species_id=species_id)
			self.populationTable.writeOneCell(oneCell, cellType=2)
			self.flush()
		return self.checkPopulation(name=name)	#would this work without flush()?
	
	def getPopulation(self, name=None, id=None, size=None, speciesName=None):
		"""
		2013.3.8
		"""
		entry = None
		if name or id:
			entry= self.checkPopulation(name=name, id=id)
			if not entry:
				entry = self.addPopulation(name=name, size=size, speciesName=speciesName)
		return entry
	
	def checkSpecies(self, name=None, id=None):
		"""
		2013.3.8
		"""
		return self.checkIfEntryInTable(tableObject=self.speciesTable,
			name=name, entry_id=id)
	
	def addSpecies(self, name=None, scientific_name=None, ploidy=None):
		"""
		2013.3.8
		"""
		if name:
			oneCell = PassingData(name=name, scientific_name=scientific_name, ploidy=ploidy)
			self.speciesTable.writeOneCell(oneCell, cellType=2)
			self.flush()
		return self.checkSpecies(name=name)	#would this work without flush()?
	
	def getSpecies(self, name=None, id=None, scientific_name=None, ploidy=None):
		"""
		2013.3.8
		"""
		entry = None
		if name or id:
			entry= self.checkSpecies(name=name, id=id)
			if not entry:
				entry = self.addSpecies(name=name,
					scientific_name=scientific_name, ploidy=ploidy)
		return entry
	
	def checkIndividual(self, name=None, id=None):
		"""
		2013.3.8
		"""
		return self.checkIfEntryInTable(tableObject=self.individualTable,
			name=name, entry_id=id)
	
	def addIndividual(self, name=None, family_id = None, father_name = None,
		mother_name = None, sex = None, phenotype = None, \
		populationName=None, speciesName=None, ploidy=None):
		"""
		2013.3.8
		"""
		if name:
			population_id = None
			species_id = None
			if speciesName:
				species = self.getSpecies(name=speciesName, ploidy=ploidy)
				if species:
					species_id = species.id
			if populationName:
				population = self.getPopulation(name=populationName, speciesName=speciesName)
				if population:
					population_id = population.id
			oneCell = PassingData(name=name,family_id =family_id, father_name=father_name,\
				mother_name=mother_name, sex=sex, phenotype=phenotype,\
				population_id=population_id)
			self.individualTable.writeOneCell(oneCell, cellType=2)
			self.flush()
			if name in self._individualName2ID:
				sys.stderr.write("Error: individual %s is not unique, already in _individualName2ID with id=%s.\n"%\
								(name, self._individualName2ID.get(name)))
				raise
			else:
				self._individualName2ID[name] = self.individualTable.no_of_rows
				#nrows is not updated until flush()
		
		#would this work without flush()?
		return self.checkIndividual(name=name)
	
	def getIndividual(self, name=None, id=None, family_id = None, father_name = None, \
					mother_name = None, sex = None, phenotype = None, \
					populationName=None, speciesName=None, ploidy=None):
		"""
		2013.3.8 query the individual table
		"""
		entry = None
		if name or id:
			entry= self.checkIndividual(name=name, id=id)
			if not entry:
				entry = self.addIndividual(name=name, family_id=family_id,
					father_name=father_name,
					mother_name=mother_name, sex=sex, phenotype=phenotype,
					populationName=populationName, speciesName=speciesName,
					ploidy=ploidy)
		return entry
	
	def checkLocus(self, name=None, id=None):
		"""
		2013.3.10
		"""
		return self.checkIfEntryInTable(tableObject=self.locusTable,
			name=name, entry_id=id)
	
	def addLocus(self, name=None, chromosomeName=None,\
		start = None, stop = None, ref_allele = None, ref_allele_length=None,
		ref_allele_frequency =None, alt_allele=None, alt_allele_length=None,\
		alt_allele_frequency=None, generation_mutation_arose=None,
		generation_mutation_fixed=None,\
		mutation_type =None, fitness = None, ancestral_amino_acid =None, \
		derived_amino_acid =None, **keywords):
		"""
		2013.3.8
		"""
		if name:
			chromosome_id = None
			if chromosomeName:
				chromosome_id = self.getChromosome(name=chromosomeName).id
			oneCell = PassingData(name=name, chromosome_id=chromosome_id,
				start = start, stop = stop, \
				ref_allele = ref_allele, ref_allele_length=ref_allele_length,\
				ref_allele_frequency =ref_allele_frequency,
				alt_allele=alt_allele, alt_allele_length=alt_allele_length,\
				alt_allele_frequency=alt_allele_frequency,
				generation_mutation_arose=generation_mutation_arose, \
				generation_mutation_fixed=generation_mutation_fixed,\
				mutation_type =mutation_type, fitness = fitness,
				ancestral_amino_acid =ancestral_amino_acid, \
				derived_amino_acid =derived_amino_acid, **keywords)
			self.locusTable.writeOneCell(oneCell, cellType=2)
			self.flush()
		return self.checkLocus(name=name)	#would this work without flush()?
	
	def getLocus(self, name=None, id=None, chromosomeName=None,\
				start = None, stop = None, ref_allele = None, ref_allele_length=None,\
				ref_allele_frequency =None, alt_allele=None, alt_allele_length=None,\
				alt_allele_frequency=None, generation_mutation_arose=None,
				generation_mutation_fixed=None,\
				mutation_type =None, fitness = None, ancestral_amino_acid =None, \
				derived_amino_acid =None, **keywords):
		"""
		2013.3.8
		"""
		entry = None
		if name or id:
			entry= self.checkLocus(name=name, id=id)
			if not entry:
				entry = self.addLocus(name=name, chromosomeName=chromosomeName,\
					start = start, stop = stop, ref_allele = ref_allele,
					ref_allele_length=ref_allele_length,\
					ref_allele_frequency =ref_allele_frequency,
					alt_allele=alt_allele, alt_allele_length=alt_allele_length,\
					alt_allele_frequency=alt_allele_frequency,
					generation_mutation_arose=generation_mutation_arose, \
					generation_mutation_fixed=generation_mutation_fixed,\
					mutation_type =mutation_type, fitness = fitness,
					ancestral_amino_acid =ancestral_amino_acid, \
					derived_amino_acid =derived_amino_acid, **keywords)
		return entry
	
	def checkPolymorphism(self, name=None, id=None):
		"""
		2013.3.10
		"""
		return self.checkIfEntryInTable(tableObject=self.polymorphismTable,
			name=name, entry_id=id)
	
	def addPolymorphism(self, name=None, individualName=None, locusName=None,
		chromosome_copy = None,\
		allele_sequence=None, allele_sequence_length=None, allele_type =None, **keywords):
		"""
		2013.3.10
		"""
		if name:
			individual_id = self.getIndividual(name=individualName).id
			locus_id = self.getLocus(name=locusName).id
			oneCell = PassingData(name=name, individual_id=individual_id, locus_id = locus_id,
				chromosome_copy=chromosome_copy,
				allele_sequence = allele_sequence,
				allele_sequence_length=allele_sequence_length,\
				allele_type=allele_type, **keywords)
			self.polymorphismTable.writeOneCell(oneCell, cellType=2)
			self.flush()
		return self.checkPolymorphism(name=name)	#would this work without flush()?
	
	def getPolymorphism(self, name=None, id=None, individualName=None, locusName=None, chromosome_copy = None,\
					allele_sequence=None, allele_sequence_length=None, allele_type =None, **keywords):
		"""
		2013.3.10
		"""
		entry = None
		if name or id:
			entry= self.checkPolymorphism(name=name, id=id)
			if not entry:
				entry = self.addPolymorphism(name=name, individualName=individualName, locusName =locusName, \
					chromosome_copy=chromosome_copy,\
					allele_sequence = allele_sequence, allele_sequence_length=allele_sequence_length,\
					allele_type=allele_type, **keywords)
		return entry
	
	def writePedigreeDiGraph2IndividualTable(self, diGraph=None, populationName=None, speciesName=None, \
											ploidy=None, **keywords):
		"""
		2013.3.7 diGraph is an instance of pymodule.algorithm.graph.DiGraphWrapper
			#. make sure all parents have an entry in the table (so that they could be referenced in recombination-table)
			#. establish _individualName2ID map so that it could be used in outputting recombination or polymorphism data
			#. sex is set to 0 (unknown). 1=male, 2=female
			#. no good for appending to an individual table with entries in it already
		"""
		sys.stderr.write("Writing the pedigree directed graph out ...")
		
		family_id = "1"
		for individualName in diGraph:
			parents = diGraph.predecessors(individualName)
			father_name = None
			mother_name = None
			if len(parents)>=1:
				father_name = parents[0]
			if len(parents)>=2:
				mother_name = parents[1]
			self.addIndividual(name=individualName, family_id=family_id, father_name=father_name, \
							mother_name=mother_name, sex=0, phenotype=None, populationName=populationName, \
							speciesName=speciesName, ploidy=ploidy)
			
		sys.stderr.write(" %s individuals.\n"%(self.individualTable.no_of_rows))
	
	def writeChrStartStopTupleList2LocusTable(self, chr_start_stop_list=None, chromosomeLength=None,\
											speciesName=None, ploidy=None):
		"""
		2013.3.7
			#. establish _locus_index2id, to be used in writeIndividualName2PolymorphismData()
			#. make sure chr_start_stop_list is in the same order as the haplotype in writeIndividualName2PolymorphismData()
		"""
		sys.stderr.write("Writing a %s-element list of (chr, start,stop) out ..."%(len(chr_start_stop_list)))
		chr_start_stop_list.sort()	#make sure it's sorted
		if ploidy is None:
			ploidy=self.ploidy
		for i in range(len(chr_start_stop_list)):
			chromosomeName, start, stop = chr_start_stop_list[i][:3]
			if chromosomeName:
				chromosomeEntry = self.getChromosome(name=chromosomeName, length=chromosomeLength, speciesName=speciesName,\
													ploidy=ploidy)
			else:
				chromosomeEntry = None
			name = '%s_%s_%s'%(chromosomeName, start, stop)
			oneCell = PassingData(name=name, chromosome_id=getattr(chromosomeEntry, 'id', None), start=start, stop=stop)
			self.locusTable.writeOneCell(oneCell, cellType=2)
			self._locus_index2id[i] = self.locusTable.no_of_rows
		sys.stderr.write("%s loci \n")
		return self._locus_index2id
	
	def writeIndividualName2PolymorphismData(self, individualName2polymorphismData=None, \
											locus_index2id=None, speciesName=None, ploidy=None):
		"""
		2013.3.7
			if locus_index2id is not available, raise error
		"""
		sys.stderr.write("Writing individualName2polymorphismData (%s individuals) out ..."%\
						(len(individualName2polymorphismData)))
		if locus_index2id is None:
			locus_index2id = self._locus_index2id
		counter = 0
		for individualName, polymorphismData in individualName2polymorphismData.items():
			individual_id = self.getIndividual(individualName, speciesName=speciesName, ploidy=ploidy).id
			for i in range(len(polymorphismData.haplotypeList)):
				haplotype = polymorphismData.haplotypeList[i]
				for j in range(len(haplotype)):
					locus_id = locus_index2id.get(j)
					if locus_id is None:
						sys.stderr.write("Error: no locus_id for locus index %s.\n"%(j))
						raise
					oneCell = PassingData(individual_id=individual_id, locus_id=locus_id,\
										chromosome_copy=i, allele_sequence=haplotype[j],\
										allele_sequence_length=len(haplotype[j]), allele_type=1)
					self.polymorphismTable.writeOneCell(oneCell, cellType=2)
					counter += 1
		sys.stderr.write(" %s alleles outputted.\n"%(counter))
		
	
	def writeRecombinationEvents(self, parentName=None, childName=None, recombinationLocationList=None):
		"""
		2013.3.7
		"""
		parent_id = self.getIndividual(parentName).id
		child_id = self.getIndividual(childName).id
		for position in recombinationLocationList:
			oneCell = PassingData(parent_id=parent_id, child_id=child_id, position=position)
			self.recombinationTable.writeOneCell(oneCell, cellType=2)
		
		
	


#!/usr/bin/env python3
"""
2012.10.8
    module related to Fasta file
"""
import os,sys
from Bio import SeqIO
from palos import ProcessOptions, PassingData

class FastaFile(object):
    __doc__ = __doc__
    option_default_dict = {
        ('inputFname', 0, ): [None, 'i', 1,
            'a VCF input file to read in the VCF content.'],\
        ('outputFname', 0, ): [None, 'o', 1, 
            'a VCF output file to hold the the VCF content.'],\
        ('debug', 0, int):[0, 'b', 0, 'toggle debug mode'],\
        ('report', 0, int):[0, 'r', 0, 'toggle report, more verbose stdout/stderr.']
        }
    
    def __init__(self, **keywords):
        """
        """
        self.ad = ProcessOptions.process_function_arguments(keywords,
            self.option_default_dict, error_doc=self.__doc__, \
            class_to_have_attr=self)
        self.inf = None
        self._initializeInput(inputFname=self.inputFname)
        self._seqTitle2sequence = None
        self._seqTitleList = None	#2013.07.08
        
        self.outf = None
    
    def _initializeInput(self, inputFname=None):
        """
        """
        if inputFname:
            if inputFname[-3:]=='.gz':
                import gzip
                self.inf = gzip.open(inputFname, 'rb')
            else:
                self.inf = open(inputFname)
        
    def close(self,):
        """
        2012.5.10
        """
        if self.inf:
            del self.inf
            
    def __iter__(self):
        """
        make itself an iterator
        """
        #or should it be SeqIO.parse(self.inf, "fasta") ?
        return self
    
    def __next__(self):
        """
        """
        try:
            return next(SeqIO.parse(self.inf, "fasta"))
        except:
            raise StopIteration
    
    @property
    def seqTitleList(self):
        """
        2013.07.08
        """
        if self._seqTitleList is None:
            self.seqTitleList = ()
        return self._seqTitleList
    
    @seqTitleList.setter
    def seqTitleList(self, argument_ls=[]):
        """
        2013.07.08
        """
        sys.stderr.write("Getting seqTitleList from %s ..."%(self.inputFname))
        if self.inf.tell()!=0:
            self.inf.seek(0)
        self._seqTitleList = []
        for record in SeqIO.parse(self.inf, "fasta"):
            seqTitle = record.id
            self._seqTitleList.append(seqTitle)
        sys.stderr.write("%s sequences.\n"%(len(self._seqTitleList)))
        
    @property
    def seqTitle2sequence(self):
        """
        2012.10.8
        """
        if self._seqTitle2sequence is None:
            self.seqTitle2sequence = ()
        return self._seqTitle2sequence
    
    @seqTitle2sequence.setter
    def seqTitle2sequence(self, argument_ls=[]):
        """
        2012.10.8
        """
        sys.stderr.write("Getting seqTitle2sequence from %s ..."%(self.inputFname))
        if self.inf.tell()!=0:
            self.inf.seek(0)
        self._seqTitle2sequence = {}
        for record in SeqIO.parse(self.inf, "fasta"):
            seqTitle = record.id
            seq = record.seq
            self._seqTitle2sequence[seqTitle] = seq
        sys.stderr.write("%s sequences.\n"%(len(self._seqTitle2sequence)))

    
    def getSequence(self, seqTitle=None, start=None, stop=None):
        """
        2012.10.8
            coordinates are 1-based and full closure.
            example: start=1, stop=10 => [1,10]
        """
        sequence = self.seqTitle2sequence.get(seqTitle)
        if sequence:
            return sequence[start-1:stop]
        else:
            return None


# -*- coding: future_fstrings -*-
#for Python2&3 compatibility
from __future__ import absolute_import, division, print_function
from builtins import (bytes, str, open, super, range,
    zip, round, input, int, pow, object)
from future import standard_library
standard_library.install_aliases()
from future.builtins import next
from future.builtins import object

import os, sys, csv, re, numpy
import logging

def dict_map(dictionaryStructure=None, ls=None, type=1):
    """
    2008-04-03 copied from annot.codense.common
    10-13-05
        add type 2 to return item itself if mapping is not available
    2006-12-21
        add type 3 to extract a smaller map
    2007-05-14
        bug, "if value" could miss 0
    """
    if type==3:
        new_list = {}
        #it's a dictionary
        for item in ls:
            value = dictionaryStructure.get(item)
            if value is not None:
                new_list[item] = value
    else:
        new_list = []
        for item in ls:
            value = dictionaryStructure.get(item)
            if value is not None:
                new_list.append(value)
            elif type==2:
                new_list.append(item)
    
    return new_list

class PassingData(object):
    """
    05/09/08
        a class to hold any data structure
    """
    
    def __init__(self, **keywords):
        """
        2008-5-12
            add keyword handling
        """
        for argument_key, argument_value in keywords.items():
            setattr(self, argument_key, argument_value)
            #setattr(self, argument_key, argument_value)
        
    def __str__(self):
        """
        2010-6-17
            a string-formatting function
        """
        return_ls = []
        for attributeName in dir(self):
            if attributeName.find('__')==0:	#ignore the 
                continue
            value = getattr(self, attributeName, None)
            return_ls.append("%s = %s"%(attributeName, value))
            
        return ", ".join(return_ls)
    
    def __getitem__(self, key):
        """
        2012.12.20 enable it to work like a dictionary
        i.e. pdata.chromosome or pdata['chromosome'] is equivalent if 
            attribute 0 is chromosome.
        """
        return self.__getattribute__(key)

class PassingDataList(list, object):
    """
    2012.12.20 could access value/attributes as a dictionary due to
         __getitem__(self, key)
    2012.11.24
    Could be accessed as a list as well.
    The position of each individual attribute in this list is based on
            the order in which the variables are put into the list.
    i.e.
        association_peak = PassingDataList()
        association_peak.chromosome=peak_start_data_obj.chromosome
        association_peak.start=peak_start_data_obj.peak_start
        association_peak.stop=intersection_point.x()
        association_peak[0] == association_peak.chromosome	#this is True
    
    But if you assign everything in the initialization, the order is not guaranteed. i.e.
        association_peak = PassingDataList(
            chromosome=peak_start_data_obj.chromosome,
            start=peak_start_data_obj.peak_start,\
            stop=intersection_point.x(), start_locus_id=1)
        association_peak[0] == association_peak.chromosome
        #not sure. the order of arguments is not preserved inside __init__(**keywords).
    
    list has to be ahead of object as parental class, otherwise
        "TypeError: Error when calling the metaclass basesCannot create a 
            consistent method resolution order (MRO) for bases list, object"
    a list needs these methods.
        http://snipplr.com/view.php?codeview&id=6626
        http://docs.python.org/2/reference/datamodel.html#sequence-types
    05/09/08
        a class to hold any data structure
    """
    
    def __init__(self, **keywords):
        """
        2012.11.24
        2008-5-12
            add keyword handling
        """
        object.__setattr__(self, 'attributeName2Index', {})
        #watch: self (list) may contains items not recorded in attributeName2Index. 
        
        for argument_key, argument_value in keywords.items():
            setattr(self, argument_key, argument_value)
            #setattr(self, argument_key, argument_value)
        
    def __str__(self):
        """
        2010-6-17
            a string-formatting function
        """
        return_ls = []
        try:
            for attributeName, index in self.attributeName2Index.items():
                value = self[index]
                #if attribute_name.find('__')==0:	#ignore the 
                #	continue
                #value = getattr(self, attributeName, None)
                return_ls.append("%s = %s"%(attributeName, value))
        except AttributeError:
            for attributeName in dir(self):
                if attributeName.find('__')==0:	#ignore the 
                    continue
                value = getattr(self, attributeName, None)
                return_ls.append("%s = %s"%(attributeName, value))
            
        return ", ".join(return_ls)
    
    def __getattribute__(self, name):
        """
        2012.11.24
        """
        try:
            returnAttr = object.__getattribute__(self, name)
        except AttributeError:
            attributeName2Index = object.__getattribute__(self, 'attributeName2Index')
            index = attributeName2Index.get(name, None)
            if index is not None:
                returnAttr =  self.__getitem__(index)
            else:
                returnAttr = None
        return returnAttr
    
    def __setattr__(self, name, value):
        """
        2012.11.24
        """
        try:
            attributeName2Index = object.__getattribute__(self, 'attributeName2Index')
            attributeName2Index[name] = len(self)
            self.append(value)
        except AttributeError:
            object.__setattr__(self, name, value)
    
    def __delattr__(self, name):
        """
        2012.11.24
        """
        try:
            attributeName2Index = object.__getattribute__(self, 'attributeName2Index')
            index = attributeName2Index.get(name, None)
            if index is not None:
                del attributeName2Index[name]
                list.__delitem__(self, index)
                #self.pop(index)
        except AttributeError:
            object.__delattr__(self, name)
    
    def __delitem__(self, indexToDelete):
        """
        2012.11.24
        
        """
        list.__delitem__(self, indexToDelete)
        try:
            attributeName2Index = object.__getattribute__(self, 'attributeName2Index')
            attributeToBeDeleted = None		
            for attributeName, index in attributeName2Index.items():
                if index==indexToDelete:
                    attributeToBeDeleted = attributeName
                    break
            if attributeToBeDeleted is not None:
                del attributeName2Index[attributeToBeDeleted]
        except AttributeError:
            pass
    
    def __getitem__(self, index_or_key):
        """
        2012.12.20 enable it to retrieve item using either index or
            attributeName (like a dictionary).
        i.e. pdata[0] or pdata['chromosome'] is equivalent if attribute 0 is
             chromosome.
        """
        if type(index_or_key)==int:
            return list.__getitem__(self, index_or_key)
        else:
            attributeName2Index = object.__getattribute__(self, 'attributeName2Index')
            index = attributeName2Index.get(index_or_key, None)
            return list.__getitem__(self, index)
        
    # Mutable sequences only, provide the Python list methods.
    #def append(self, item):
    #	pass
    
    def count(self, item):
        list.count(self, item)
    
    def index(self,item):
        list.index(self, item)
    
    #def extend(self,other):
    #	pass
    
    def insert(self,item):
        """
        disable
        """
        pass
    
    def pop(self, index=None):
        """
        """
        if index is None:
            index = self.__len__()-1	#the last one by default
        self.__delitem__(indexToDelete=index)
    
    def remove(self,item):
        """	
        disable	
        """
        pass
    
    def reverse(self):
        """
        disable
        """
        pass
    
    def sort(self):
        """
        disable
        """
        pass
    
def importNumericArray():
    """
    2008-07-09
        numarray doesn't have int128
    2008-05-18
        give same numpy types (int, int8 ...) to other numeric modules
    2008-05-18
        add "import array as num"
        should put ImportError in except. but whatever
    2008-05-11
        import whatever available array module
    """
    import numpy as num
    """
    #old numarray and Numeric
    numpy_type2other_ls = ['int', 'int8', 'int16', 'int32', 'int64']
    for numpy_type in numpy_type2other_ls:	#make sure it has same type names
        numpy_type_in_other = numpy_type[0].upper() + numpy_type[1:]
        setattr(num, numpy_type, getattr(num, numpy_type_in_other))
    """
    return num

def figureOutDelimiter(input_fname, report=0,
    delimiter_choice_ls = ['\t', ',', ' '], use_sniff=False):
    """
    #2012.8.10 max count determines the delimiter
    2012.5.8
        if input_fname is a file object, don't delete it (deleting closes it) 
            and seek to the beginning of the file.
            bugfix: the file object could be file or gzip.GzipFile 
    2008-01-08
        don't use cs.sniff unless the user specifies it. sniff gives you
            unexpected delimiter when it's a single-column.
    2008-08-28
        nothing weird on hpc-cmb. it's a bug in other code.
        back to 'return None' if input_fname escapes all condition checking.
    2008-08-28
        try 'open(input_fname)' anyway if input_fname escapes all condition
            checking.
        something weird happened during a mpi job on hpc-cmb. the file is there.
            but escape the first condition.
    2008-05-25
        now 3 possible types of input_fname
        1. a file name (path)
        2. input_fname is a file object
        3. input_fname is input data, string
        
        for a file object or input file name:
        it could be binary file which doesn't have readline(). have to use this
            dumb approach due to '\n' might mess up sniff().
    2008-05-21
        csv.Sniffer is handy, use it figure out csv.Sniffer instead.
    2008-05-12
        try tab first
    """
    if report:
        import sys
        sys.stderr.write("Figuring out delimiter for %s ..."%input_fname)
    cs = csv.Sniffer()
    inputIsFileObject = False
    import gzip
    if isinstance(input_fname, str) and os.path.isfile(input_fname):
        inf = openGzipFile(input_fname)
    elif hasattr(input_fname, 'read') or isinstance(input_fname, gzip.GzipFile):
        #a file/gzip-file object
        inf = input_fname
        inputIsFileObject = True
    elif isinstance(input_fname, str) and not os.path.isfile(input_fname):
        #it's the input
        from io import StringIO
        inf = StringIO(input_fname)
    else:
        import sys
        logging.warn(f"{input_fname} is neither a file name nor a file object. "
            "But try 'open' it anyway.")
        inf = openGzipFile(input_fname)
    if getattr(inf, 'readline', None) is not None and use_sniff:
        #2008-01-08 don't use cs.sniff unless the user specifies it. 
        #	sniff gives you unexpected delimiter when it's a single-column.
        line = inf.readline()
        delimiter_chosen = cs.sniff(line).delimiter
    else:
        line = inf.read(20000)
        ##binary file doesn't have readline().
        # have to use this dumb approach due to '\n' might mess up sniff()
        delimiter_chosen = None
        delimiterData = PassingData(delimiterWithMaxCount='\t', maxCount=-1)
        for delimiter in delimiter_choice_ls:
            delimiter_count = line.count(delimiter)
            if delimiter_count>delimiterData.maxCount:
                delimiterData.delimiterWithMaxCount = delimiter
                delimiterData.maxCount = delimiter_count
                
        delimiter_chosen = delimiterData.delimiterWithMaxCount
    if inputIsFileObject:
        inf.seek(0)
    else:
        del inf
    if report:
        sys.stderr.write("Done.\n")
    return delimiter_chosen

def get_gene_symbol2gene_id_set(curs, tax_id, table='genome.gene_symbol2id',
    upper_case_gene_symbol=0):
    """
    2008-07-10 derived from annot.bin.codense.common.get_gene_symbol2gene_id()
    """
    sys.stderr.write("Getting gene_symbol2gene_id_set...")
    gene_symbol2gene_id_set = {}
    curs.execute("select gene_id, gene_symbol from %s where tax_id=%s"%(table, tax_id))
    rows = curs.fetchall()
    for row in rows:
        gene_id, gene_symbol = row
        if upper_case_gene_symbol:
            gene_symbol = gene_symbol.upper()
        if gene_symbol not in gene_symbol2gene_id_set:
            gene_symbol2gene_id_set[gene_symbol] = set()
        gene_symbol2gene_id_set[gene_symbol].add(gene_id)
    sys.stderr.write(" %s entries. Done.\n"%len(gene_symbol2gene_id_set))
    return gene_symbol2gene_id_set

def get_gene_id2gene_symbol(curs, tax_id, table='genome.gene', upper_case_gene_symbol=0):
    """
    2008-11-14
        reverse of get_gene_symbol2gene_id_set
    """
    sys.stderr.write("Getting gene_id2gene_symbol ...")
    gene_id2gene_symbol = {}
    curs.execute("select gene_id, gene_symbol from %s where tax_id=%s"%(table, tax_id))
    rows = curs.fetchall()
    for row in rows:
        gene_id, gene_symbol = row
        if upper_case_gene_symbol:
            gene_symbol = gene_symbol.upper()
        if gene_id not in gene_id2gene_symbol:
            gene_id2gene_symbol[gene_id] = gene_symbol
        else:
            logging.warn(f"Warning: gene_id {gene_id}({gene_symbol}) already "
            f"in gene_id2gene_symbol with symbol={gene_id2gene_symbol[gene_id]}.")
    sys.stderr.write(" %s entries. Done.\n"%len(gene_id2gene_symbol))
    return gene_id2gene_symbol

class FigureOutTaxID(object):
    __doc__ = """
    2012.8.28 deprecated. moved to pymodule/TaxonomyDB.py
    2008-07-29 class to figure out tax_id using postgres database taxonomy schema
    """
    option_default_dict = {
        ('hostname', 1, ): ['localhost', 'z', 1, 'hostname of the db server', ],\
        ('dbname', 1, ): ['graphdb', 'd', 1, 'database name', ],\
        ('schema', 1, ): ['taxonomy', 'k', 1, 'database schema name', ],\
        ('db_user', 0, ): [None, 'u', 1, 'database username', ],\
        ('db_passwd', 0, ): [None, 'p', 1, 'database password', ],\
    }
    def __init__(self, hostname='localhost', dbname='graphdb',
        schema='taxonomy', db_user=None, db_passwd=None):
        """
        2008-07-29
        """
        self.hostname = hostname
        self.dbname = dbname
        self.schema = schema
        self.db_user = db_user
        self.db_passwd = db_passwd
    
    @property
    def curs(self):
        from db import db_connect
        conn, curs =  db_connect(self.hostname, self.dbname, self.schema,
            user=self.db_user, password=self.db_passwd)
        return curs
    
    @property
    def scientific_name2tax_id(self):
        """
        2012.6.6
            update it to get table names from TaxonomyDB
        """
        from palos.db import TaxonomyDB
        scientific_name2tax_id = {}
        curs = self.curs
        curs.execute(f"SELECT n.name_txt, n.tax_id FROM "
            f"taxonomy.{TaxonomyDB.Name.tablename} n, "
            f"taxonomy.{TaxonomyDB.Node.tablename} o "
            f"where n.name_class='scientific name' "
            f"and n.tax_id=o.tax_id and o.rank='species'")
        rows = curs.fetchall()
        for row in rows:
            scientific_name, tax_id = row
            scientific_name2tax_id[scientific_name] = tax_id
        return scientific_name2tax_id
    
    def returnTaXIDGivenScientificName(self, scientific_name):
        return self.scientific_name2tax_id.get(scientific_name)
    
    def returnTaxIDGivenSentence(self, sentence):
        """
        2008-07-29
        """
        tax_id_to_return = None
        for scientific_name, tax_id in self.scientific_name2tax_id.items():
            if sentence.find(scientific_name)>=0:
                tax_id_to_return = tax_id
                break
        return tax_id_to_return

def getColName2IndexFromHeader(header, skipEmptyColumn=False):
    """
    2011-2-11
        add argument skipEmptyColumn
    2008-09-16
        convenient function to read input files with flexible column order.
        One variable doesn't have to be in the same column in different files,
         as far as the name is same.
    """
    col_name2index = {}
    for i in range(len(header)):
        column_name = header[i]
        if skipEmptyColumn and not column_name:
            #skips empty column
            continue
        col_name2index[column_name] = i
    return col_name2index

def getListOutOfStr(list_in_str=None, data_type=int, separator1=',', 
    separator2='-'):
    """
        This function parses a list from a string representation of a list, 
            such as '1,3-7,11'=[1,3,4,5,6,7,11].
    If only separator2, '-', is used ,all numbers have to be integers.
    If all are separated by separator1, it could be in non-int data_type.
    strip the strings as much as u can.
    if separator2 is None or nothing or 0, it wont' be used.

    Examples:
        self.chromosomeList = utils.getListOutOfStr('1-5,7,9', data_type=str,
            separator2=None)
    """
    list_to_return = []
    if list_in_str=='' or list_in_str is None:
        return list_to_return
    list_in_str = list_in_str.strip()
    if list_in_str=='' or list_in_str is None:
        return list_to_return
    if type(list_in_str)==int:
        #just one integer, put it in and return immediately
        return [list_in_str]
    index_anchor_ls = list_in_str.split(separator1)
    for index_anchor in index_anchor_ls:
        index_anchor = index_anchor.strip()
        if len(index_anchor)==0:
            continue
        if separator2:
            start_stop_tup = index_anchor.split(separator2)
        else:
            start_stop_tup = [index_anchor]
        if len(start_stop_tup)==1:
            list_to_return.append(data_type(start_stop_tup[0]))
        elif len(start_stop_tup)>1:
            start_stop_tup = list(map(int, start_stop_tup))
            list_to_return += list(range(start_stop_tup[0], start_stop_tup[1]+1))
    list_to_return = list(map(data_type, list_to_return))
    return list_to_return

def getStrOutOfList(ls, separator=','):
    """
    #2013.2.24 bugfix
    2012.11.25 reverse of getListOutOfStr
    """
    #2013.2.24 bugfix
    if not ls:
        return ''
    firstElement = ls[0]
    if type(firstElement)!=str:
        ls_str = list(map(str, ls))
    else:
        ls_str = ls
    return separator.join(ls_str)

def getSuccinctStrOutOfList(ls=None, step=1, separator=','):
    """
    examples:
        utils.getSuccinctStrOutOfList(phenotype_method_id_of_associations_set)
        utils.getSuccinctStrOutOfList(phenotype_method_id_ls)
    
    2013.2.6 ls could be set as well
    2013.1.28 bugfix, ls could be empty list. return ''
    2012.12.21 use , and - to represent a list of numbers in fewest possible characters
        step controls the order of numbers.
        separator separates between different spans or singletons.
        like 1,2,3,4,7 => 1-4,7.
    """
    import copy
    ls_copy = copy.deepcopy(list(ls))
    ls_copy.sort()
    spanStartValue = None
    #ls_copy[0]
    spanStopValue = None
    span_tuple_ls = []
    for i in range(len(ls_copy)):
        if spanStartValue is None:
            spanStartValue = ls_copy[i]
            spanStopValue = spanStartValue
        else:
            previousValue = ls_copy[i-1]
            newValue = ls_copy[i]
            if newValue == (previousValue + step):
                spanStopValue = newValue
                #increase the spanStopValue
            else:
                #more than the step
                spanStopValue = previousValue
                if spanStartValue==spanStopValue:	#singleton
                    span_tuple_ls.append(str(spanStartValue))
                else:
                    span_tuple_ls.append('%s-%s'%(spanStartValue, spanStopValue))
                spanStartValue = newValue
                spanStopValue = newValue
    if ls_copy:
        #handle the last span
        if spanStopValue is None and ls_copy:
            spanStopValue = ls_copy[-1]
        if spanStartValue==spanStopValue:	#singleton
            span_tuple_ls.append(str(spanStartValue))
        else:
            span_tuple_ls.append('%s-%s'%(spanStartValue, spanStopValue))
    return separator.join(span_tuple_ls)

def runLocalCommand(commandline, report_stderr=True, report_stdout=False):
    """
    2011.12.19
        output stdout/stderr only when there is something to output
    2008-1-5
        copied from utility/grid_job_mgr/hpc_cmb_pbs.py
    2008-11-07
        command_handler.communicate() is more stable than
            command_handler.stderr.read()
    2008-11-04
        refactor out of runRemoteCommand()
        
        run a command local (not on the cluster)
    """
    import subprocess
    from io import StringIO
    command_handler = subprocess.Popen(commandline, shell=True,
        stderr=subprocess.PIPE, stdout=subprocess.PIPE,
        encoding='utf8')
    #command_handler.wait() #Warning: This will deadlock if the child process
    #   generates enough output to a stdout or stderr pipe
    #	such that it blocks waiting for the OS pipe buffer to accept more data.
    # Use communicate() to avoid that.
    
    #command_handler.stderr.read() command_handler.stdout.read() also constantly
    #   deadlock the whole process.
    
    stderr_content = None
    stdout_content = None
    stdout_content, stderr_content = command_handler.communicate()
    #to circumvent deadlock caused by command_handler.stderr.read()
    output_stdout = None
    output_stderr = None
    if not report_stdout:
        #if not reporting, assume the user wanna to have a file handler returned
        output_stdout = StringIO(stdout_content)
    if not report_stderr:
        output_stderr = StringIO(stderr_content)
    
    if report_stdout and stdout_content:
        sys.stderr.write('stdout of %s: %s \n'%(commandline, stdout_content))
    
    if report_stderr and stderr_content:
        sys.stderr.write('stderr of %s: %s \n'%(commandline, stderr_content))
    
    return_data = PassingData(commandline=commandline, 
        output_stdout=output_stdout, output_stderr=output_stderr,
        stderr_content=stderr_content, stdout_content=stdout_content)
    return return_data


#2009-2-4 code to link something like AT1G01010.1 or AT1G01040 to NCBI gene id
#refactored out of PutGeneListIntoDB.py
p_acc_ver = re.compile(r'(\w+)\.(\d+)')
#2008-12-11 only alphanumeric characters in gene_symbol
#  (suzi's file contains weird characters sometimes)	
p_acc = re.compile(r'(\w+)')
def getGeneIDSetGivenAccVer(acc_ver, gene_symbol2gene_id_set, noUpperCase=0):
    if not noUpperCase:
        gene_symbol = acc_ver.upper()
    if p_acc_ver.search(gene_symbol):
        gene_symbol, version = p_acc_ver.search(gene_symbol).groups()
    if p_acc.search(gene_symbol):
        #2008-12-11 pick out alphanumeric characters
        gene_symbol, = p_acc.search(gene_symbol).groups()
    gene_id_set = gene_symbol2gene_id_set.get(gene_symbol)
    return gene_id_set


def calGreatCircleDistance(lat1=None, lon1=None, lat2=None, lon2=None,
    earth_radius=6372.795):
    """
    2009-4-18
        copied from CreatePopulation.cal_great_circle_distance()
        distance in km
    2007-06-17 copied from 2007-07-11
    http://en.wikipedia.org/wiki/Great-circle_distance
    """
    import math
    lat1_rad = lat1*math.pi/180
    lon1_rad = lon1*math.pi/180
    lat2_rad = lat2*math.pi/180
    lon2_rad = lon2*math.pi/180
    long_diff = abs(lon1_rad-lon2_rad)
    sin_lat1 = math.sin(lat1_rad)
    cos_lat1 = math.cos(lat1_rad)
    sin_lat2 = math.sin(lat2_rad)
    cos_lat2 = math.cos(lat2_rad)
    spheric_angular_diff = math.atan2(
        math.sqrt(math.pow(cos_lat2*math.sin(long_diff),2) + 
        math.pow(cos_lat1*sin_lat2-sin_lat1*cos_lat2*math.cos(long_diff),2)),
        sin_lat1*sin_lat2+cos_lat1*cos_lat2*math.cos(long_diff))
    return earth_radius*spheric_angular_diff
    
def addExtraToFilenamePrefix(filename, extra):
    """
    2010-4-5
    add the extra bits (string, integer, etc.) before the file name suffix
    """
    import os
    fname_prefix, fname_suffix = os.path.splitext(filename)
    fname_prefix += '_%s'%extra
    return fname_prefix + fname_suffix

def addExtraLsToFilenamePrefix(filename, extra_ls):
    """
    2010-5-9
    add a list of the extra bits (string, integer, etc.)
        before the file name suffix
    """
    import os
    for extra in extra_ls:
        filename = addExtraToFilenamePrefix(filename, extra)
    return filename

def returnAnyValueIfNothing(string, data_type=int, defaultValue=0):
    """
    2010-12-15
    used in Transfac.src.GeneASNXML2gene_mapping.return_datetime() in case 
        nothing is returned.
    """
    if string:
        return data_type(string)
    else:
        return defaultValue


def Denary2Binary(n):
    '''
    2011-2-9
        convert denary integer n to binary string bStr
        
        copied from http://www.daniweb.com/code/snippet216539.html
        
        # convert a decimal (denary, base 10) integer to a binary string (base 2)
        # tested with Python24   vegaseat	6/1/2005
    '''
    bStr = ''
    if n < 0:
        raise ValueError("must be a positive integer")
    if n == 0: return '0'
    while n > 0:
        bStr = str(n % 2) + bStr
        n = n >> 1
    return bStr

def int2bin(n, count=24):
    """
    2011-2-9
        opposite of Denary2Binary(), same as int(binaryStr, 2)
        
        copied from http://www.daniweb.com/code/snippet216539.html
    
    returns the binary of integer n, using count number of digits
    
    """
    return "".join([str((n >> y) & 1) for y in range(count-1, -1, -1)])

"""
# this test runs when used as a standalone program, but not as an imported module
# let's say you save this module as den2bin.py and use it in another program
# when you import den2bin the __name__ namespace would now be  den2bin  and the
# test would be ignored
if __name__ == '__main__':
    print Denary2Binary(255)  # 11111111
    
    # convert back to test it
    print int(Denary2Binary(255), 2)  # 255
    
    print
    
    # this version formats the binary
    print int2bin(255, 12)  # 000011111111
    # test it
    print int("000011111111", 2)  # 255
    
    print

    # check the exceptions
    print Denary2Binary(0)
    print Denary2Binary(-5)  # should give a ValueError
"""

def processRegexpString(p_str):
    """
    2011-4-30
        copied from a pylons controller. used to process regular expressions 
        passed from web client.
    2009-4-3
        if p_str includes '(' or ')', mysql complains: ERROR 1139 (42000):
             Got error 'parentheses not balanced' from regexp
    """
    p_str = p_str.replace('(', '\(')
    #python re module only needs one '\'
    p_str = p_str.replace(')', '\)')
    p_str_sql = p_str.replace('(', '\\\(')
    #mysql needs more
    p_str_sql = p_str_sql.replace(')', '\\\)')
    return PassingData(p_str=p_str, p_str_sql=p_str_sql)


def sortCMPBySecondTupleValue(a, b):
    """
    2011-3-29
        a and b are list or tuple
    """
    return cmp(a[1], b[1])

def sshTunnel(serverHostname="dl324b-1.cmb.usc.edu", port="5432",
    middleManCredential="polyacti@login3"):
    """
    2011-9-5
    replace runLocalCommand() with os.system()
    runLocalCommand() calls command_handler.communicate() which causes
        the program to get stuck
        might be caused by "ssh -f" daemon behavior.
    Correct way of replacing os.system() through Popen():
        sts = os.system("mycmd" + " myarg")
        ==>
        p = Popen("mycmd" + " myarg", shell=True)
        sts = os.waitpid(p.pid, 0)[1]
        
    2011-8-15
    through middleManCredential, run a ssh tunnel to allow access to
        serverHostname:port as localhost:port
    
    example:
        # forward postgresql db on dl324b-1 to localhost
        sshTunnel("dl324b-1.cmb.usc.edu", 5432, "polyacti@login3")
        
    """
    commandline = "ssh -N -f -L %s:%s:%s %s"%(port, serverHostname, port,
        middleManCredential)
    #2011-9-5 uncomment following. program will get stuck. might be caused
    #    by "ssh -f" daemon behavior
    #runLocalCommand(commandline, report_stderr=True, report_stdout=True)
    return os.system(commandline)

def converSolexaScoreToPhred(solexaChar):
    """
    2011-8-15
    main doc: http://en.wikipedia.org/wiki/FASTQ_format
    
    simple & approximate formula would just be ord(solexaChar)-64.
    
    full formula is at the last line of http://maq.sourceforge.net/fastq.shtml,
        which is used here.
        a slight modification: here uses log10, rather than natural log.
    """
    import math
    return 10*math.log10(1 + math.pow(10, (ord(solexaChar) - 64) / 10.0))

def getFileBasenamePrefixFromPath(path=None,
    fakeSuffixSet = set(['.gz', '.zip', '.bz2', '.bz'])):
    """
    i.e. 
        path= 'folderHighCoveragePanel/Scaffold97.unphased_familySize3.bgl'
    
        getFileBasenamePrefixFromPath(path) == 
            "Scaffold97.unphased_familySize3"
    
    it will also get rid of suffix that are in the fakeSuffixSet.
        path = "folder/input.vcf.gz"
        
        getFileBasenamePrefixFromPath(path) == "input"
    
    2013.11.24 call getRealPrefixSuffix() for actual work
    2013.06.21 convenient function
    """
    fileBasename = os.path.basename(path)
    return getRealPrefixSuffix(fileBasename,
        fakeSuffixSet=fakeSuffixSet)[0]
    
    

def getRealPrefixSuffix(path, fakeSuffix='.gz',
    fakeSuffixSet = set(['.gz', '.zip', '.bz2', '.bz'])):
    """
    The purpose of this function is to get the prefix, suffix of a 
        filename regardless of whether it has two suffices (gzipped) or one. 
    i.e.
        A file name is either sequence_628BWAAXX_4_1.fastq.gz or
             sequence_628BWAAXX_4_1.fastq (without gz).
        This function returns ('sequence_628BWAAXX_4_1', '.fastq')

    "." is considered part of the filename suffix.
    """
    fname_prefix, fname_suffix =  os.path.splitext(path)
    if fakeSuffix and fakeSuffix not in fakeSuffixSet:
        fakeSuffixSet.add(fakeSuffix)
    while fname_suffix in fakeSuffixSet:
        fname_prefix, fname_suffix =  os.path.splitext(fname_prefix)
    return fname_prefix, fname_suffix

def getAllFiles(inputDir, inputFiles=[]):
    """
    2011-9-11
        copied from file_batch_move.py
    2011-8-3
        recursively going through the directory to get all files
        
    """
    import os
    for inputFname in os.listdir(inputDir):
        #get the absolute path
        inputFname = os.path.join(inputDir, inputFname)
        if os.path.isfile(inputFname):
            inputFiles.append(inputFname)
        elif os.path.isdir(inputFname):
            getAllFiles(inputFname, inputFiles)

def sumOfReciprocals(n):
    """
    2011-10-21
        for normalized nucleotide diversity
        \pi = no-of-polymorphic-loci/sumOfReciprocals
    """
    sum = 0.0
    for i in range(n-1):
        sum = sum + 1/(i+1.0)
    return sum

def get_md5sum(filename, md5sum_command = 'md5sum'):
    """
    20200120 python3, convert md5sum from bytes (binary) to utf-8.
    2012.1.27
        copied from variation/src/Array2DB_250k.py
    """
    import subprocess
    md5sum_p = subprocess.Popen([md5sum_command, filename], 
        stderr=subprocess.PIPE, stdout=subprocess.PIPE)
    md5sum_stdout_out = md5sum_p.stdout.read()
    md5sum_stderr_out = md5sum_p.stderr.read()
    if md5sum_stderr_out:
        sys.stderr.write("%s %s failed with stderr: %s.\n"%(md5sum_command,
            filename, md5sum_stderr_out))
        sys.exit(4)
    else:
        return md5sum_stdout_out.split()[0].decode("utf-8")

def getDateStampedFilename(filename):
    """
    2012.3.26
        copied from variation.src.Stock_250kDB.ResultsMethod
    2012.3.21
        xxx.tsv => xxx.2012_3_21.tsv
    """
    from datetime import datetime
    lastModDatetime = datetime.fromtimestamp(os.stat(filename).st_mtime)
    prefix, suffix = os.path.splitext(filename)
    newFilename = '%s.%s_%s_%s%s'%(prefix, lastModDatetime.year,
        lastModDatetime.month, lastModDatetime.day, suffix)
    return newFilename

def openGzipFile(inputFname, mode='r'):
    """
    Pass encoding='utf-8' to gzip.open().
    support mode 'r', 'a', 'w'.
    If suffix is .gz, use gzip to open it
    """
    fname_suffix = os.path.splitext(inputFname)[1]
    if fname_suffix=='.gz':
        import gzip
        # encoding='utf-8' only supported in text mode.
        # binary mode returns bytes object.
        if mode=='r':
            mode='rt'
        elif mode=='w':
            mode='wt'
        elif mode=="a":
            mode = 'at'
        else:
            mode='rt'
        inf = gzip.open(inputFname, mode, encoding='utf-8')
        inf.is_gzip = True
    else:
        inf = open(inputFname, mode)
        #inf.is_gzip = False
    return inf

def comeUpSplitFilename(outputFnamePrefix=None, suffixLength=3, fileOrder=0,
    filenameSuffix=""):
    """
    '%0*d'%(suffixLength, fileOrder) is same as str(fileOrder).zfill(suffixLength).
    If fileOrder's length is beyond suffixLength,
        then it's just fileOrder itself without truncation.
    Like 001, 002, 999, 1234.
    """
    return '%s%0*d%s'%(outputFnamePrefix, suffixLength, fileOrder, filenameSuffix)

def findFilesWithOneSuffixRecursively(inputDir='./', suffix='.bam'):
    """
    If suffix is empty string, it'll get all files.
    """
    import fnmatch
    
    matches = []
    for root, dirnames, filenames in os.walk(inputDir):
        for filename in fnmatch.filter(filenames, '*%s'%(suffix)):
            matches.append(os.path.join(root, filename))
    return matches


def getFolderSize(inputDir = '.'):
    """
    """
    total_size = 0
    for dirpath, dirnames, filenames in os.walk(inputDir):
        for f in filenames:
            fp = os.path.join(dirpath, f)
            total_size += os.path.getsize(fp)
    return total_size

def getFileOrFolderSize(path='.'):
    """
    """
    file_size = None
    if path:
        if os.path.isfile(path):
            statinfo = os.stat(path)
            file_size = statinfo.st_size
        elif os.path.isdir(path):
            file_size = getFolderSize(path)
    return file_size

def getNoOfLinesInOneFileByWC(inputFname=None):
    """
    2012.7.30
    call wc to finish the task
    "wc -l"'s output looks like this:
    
    crocea@crocea:~$ wc -l shell/countNoOfVariantsInOneVCFFolder.sh
    16 shell/countNoOfVariantsInOneVCFFolder.sh
    """
    #commandline = "wc -l %s|awk -F ' ' '{print $1}'"%inputFname
    commandline = "wc -l %s"%inputFname	#just wc
    return_data = runLocalCommand(commandline, report_stderr=False, 
        report_stdout=False)
    stdout_content = return_data.stdout_content
    noOfLines = int(stdout_content.split()[0].strip())
    return noOfLines


def getNoOfLinesInOneFileByOpen(inputFname=None):
    """
    2012.7.30
        open the file and count
    """
    counter = 0
    inf = openGzipFile(inputFname)
    for line in inf:
        counter += 1
    return counter

def copyFile(srcFilename=None, dstFilename=None, copyCommand="cp -aprL",
    srcFilenameLs=None, dstFilenameLs=None):
    """
    2012.7.18
        -L of cp meant "always follow symbolic links in SOURCE".
    """
    #move the file
    commandline = '%s %s %s'%(copyCommand, srcFilename, dstFilename)
    return_data = runLocalCommand(commandline, report_stderr=True,
        report_stdout=True)
    if srcFilenameLs is not None:
        srcFilenameLs.append(srcFilename)
    if dstFilenameLs is not None:
        dstFilenameLs.append(dstFilename)
    if return_data.stderr_content:
        #something wrong. abort
        sys.stderr.write("commandline %s failed: %s\n"%(commandline,
            return_data.stderr_content))
        exitCode = 3
    else:
        exitCode = 0
    return exitCode

def getNoOfUnitsNeededToCoverN(N=None, s=None, o=None):
    """
    2012.8.25
        purpose is to figure out how many blocks/units it needs to split 
            N objects into given N.
        N could also be smaller than s and o (first block).
        
        s = noOfSitesPerUnit
        o = noOfOverlappingSites
        N = noOfTotalSites
        n = noOfUnits
        s*n - o(n-1) = N
            => n = (N-o)/(s-o)
        1. when N <= o, then use N straight.
        2. same thing for s and o.
    """
    import math
    if N<=o:
        numerator = N
    else:
        numerator = N-o
    if s<=o:
        denominator = s
    else:
        denominator = s-o
    #make sure its bigger than 1.
    noOfUnits = max(1, math.ceil(numerator/float(denominator)))
    return int(noOfUnits)

def addObjectAttributeToSet(objectVariable=None, attributeName=None,
    setVariable=None):
    """
    2012.11.22
        do not add an attribute to the set if it's not available or if it's none
    """
    attributeValue = getattr(objectVariable, attributeName, None)
    if attributeValue is not None and setVariable is not None:
        setVariable.add(attributeValue)
    return setVariable

def addObjectListAttributeToSet(objectVariable=None, attributeName=None,
    setVariable=None, data_type=int):
    """
    2012.12.2 bugfix
    2012.11.23
    """
    attributeValue = getattr(objectVariable, attributeName, None)
    flag = False
    if type(attributeValue)==numpy.ndarray:
        #"if attributeValue" fails for numpy array
        if hasattr(attributeValue, '__len__') and attributeValue.size>0:
            flag = True
    elif attributeValue or attributeValue == 0:
        flag = True
    if flag and setVariable is not None:
        if type(attributeValue)==list or type(attributeValue)==tuple or\
            type(attributeValue)==numpy.ndarray or type(attributeValue)==set:
            attributeValueList = attributeValue
        elif type(attributeValue)==str:
            attributeValueList = getListOutOfStr(attributeValue, \
                data_type=data_type, separator1=',', separator2='-')
        else:
            attributeValueList = getListOutOfStr(attributeValue, \
                data_type=data_type, separator1=',', separator2='-')
        setVariable |= set(list(attributeValueList))
    return setVariable

def pauseForUserInput(message="",
    continueAnswerSet=set(['Y', 'y', '', 'yes', 'Yes']), exitType=1):
    """
    Examples:
        #exit pending user answer
        message = "Error: plink merge file %s already exists."
            "Overwrite it? (if its associated workflows do not use it anymore.)"
            " [Y/n]:"%(outputFname)
        utils.pauseForUserInput(message=message,
            continueAnswerSet=set(['Y', 'y', '', 'yes', 'Yes']), exitType=1)
    
        #exit regardless
        message = "ERRROR in "
            "AbstractVCFWorkflow.concatenateIntervalsIntoOneVCFSubWorkflow():"
            " %s is None."%(fileBasenamePrefix)
        utils.pauseForUserInput(message=message, continueAnswerSet=None, exitType=1)
    
    
    2013.07.17
        exitType
            1: exit non-zero
            2: raise
            3: pass (ignore)
    """
    userAnswer = input("\n\t %s"%(message))
    if continueAnswerSet is not None and userAnswer in continueAnswerSet:
        pass
    elif continueAnswerSet is None:
        logging.error(f"pauseForUserInput(): continueAnswerSet is None. "
            f"Exit regardless of user answer ({userAnswer}).")
        if exitType==2:
            raise
        elif exitType==1:
            sys.exit(3)
        else:
            pass
    else:
        logging.error(f"pauseForUserInput(): user answered {userAnswer}, "
            f"interpreted as no.")
        if exitType==2:
            raise
        elif exitType==1:
            sys.exit(2)
        else:
            pass

#2012.10.5 copied from VervetDB.py
#used in getattr(individual_site_id_set, '__len__', returnZeroFunc)()
returnZeroFunc = lambda: 0

if __name__ == '__main__':
    FigureOutTaxID_ins = FigureOutTaxID()
    print(FigureOutTaxID_ins.returnTaxIDGivenSentence(
        ">gi|172045488|ref|NW_001867254.1| Physcomitrella patens subsp. "
        "patens PHYPAscaffold_10696"))



#!/usr/bin/env python3
"""
2013.05.03 a child class of MatrixFile. used to describe pedigree file in plink format (delimiter could be ' ' or tab), which looks like:
    The header line does not exist. Six columns are: FamilyID IndividualID PaternalID MaternalID Sex(1=male; 2=female; other=unknown) Phenotype.
    
    F0_1 990_699_1984014_GA_vs_524copy1 0 0 2 1
    F0_1 1513_641_1986014_GA_vs_524copy1 0 0 1 1
    F0_1 984_693_1996027_GA_vs_524copy1 1513_641_1986014_GA_vs_524copy1 990_699_1984014_GA_vs_524copy1 1 1
    F1_1 1582_1672_1993040_GA_vs_524copy1 0 0 1 1
    F1_1 1917_2966_1992045_GA_vs_524copy1 0 0 2 1
    F1_1 1931_2980_2000040_GA_vs_524copy1 1582_1672_1993040_GA_vs_524copy1 1917_2966_1992045_GA_vs_524copy1 1 1

Example:

    plinkPedigreeFile = PlinkPedigreeFile(path=self.pedigreeFname)
    familyStructureData = plinkPedigreeFile.getFamilyStructurePlinkWay()
    
    plinkPedigreeFile = PlinkPedigreeFile(path=self.pedigreeFname)
    for nodeID in plinkPedigreeFile.pedigreeGraph:
        ...
    plinkPedigreeFile.close()

"""

import sys, os, math
import copy
import networkx as nx
from palos.ProcessOptions import  ProcessOptions
from palos import utils, PassingData
from palos.algorithm import graph
from palos.io.MatrixFile import MatrixFile

class PlinkPedigreeFile(MatrixFile):
    __doc__ = __doc__
    option_default_dict = copy.deepcopy(MatrixFile.option_default_dict)
    option_default_dict.update({
        ('dummyIndividualNamePrefix', 0, ): ['dummy', '', 1,
            'the prefix to name a dummy parent (TrioCaller format). The suffix is its order among all dummies.'],\
            
        })
    def __init__(self, path=None, **keywords):
        MatrixFile.__init__(self, path=path, **keywords)
        
        self.familyID2MemberList= {}
        self.familySize2SampleIDList = {}
        self._pedigreeGraph = None
        self._childNodeSet = None
    
    @property
    def pedigreeGraph(self):
        """
        """
        if self._pedigreeGraph is None:
            self.pedigreeGraph = None
        return self._pedigreeGraph
    
    @pedigreeGraph.setter
    def pedigreeGraph(self, argument_ls=[]):
        """
        """
        graphData = self.getPedigreeGraph()
        self._pedigreeGraph = graphData.DG
        self._childNodeSet = graphData.childNodeSet
    
    def getPedigreeGraph(self,):
        """
        """
        sys.stderr.write("Getting pedigree graph from this file %s ..."%(self.path))
        
        self._resetInput()
        #self.constructColName2IndexFromHeader()	#there is no header.
        DG = graph.DiGraphWrapper()
        childNodeSet = set()
        counter = 0
        for row in self:
            DG.add_node(row.individualID)
            #in case this guy has no parents, then won't be added via add_edge()
            childNodeSet.add(row.individualID)
            if row.paternalID!='0' and row.paternalID.find(self.dummyIndividualNamePrefix)!=0:
                DG.add_edge(row.paternalID, row.individualID)
            if row.maternalID!='0' and row.maternalID.find(self.dummyIndividualNamePrefix)!=0:
                DG.add_edge(row.maternalID, row.individualID)
            counter += 1
        sys.stderr.write("%s children, %s nodes. %s edges. %s connected components.\n"%(\
            len(childNodeSet), DG.number_of_nodes(), DG.number_of_edges(), \
            nx.number_connected_components(DG.to_undirected())))
        return PassingData(DG=DG, childNodeSet=childNodeSet)
    
    def getFamilyStructure(self):
        """
        2013.07.19
        """
        sys.stderr.write("Finding unique pairs (singletons or groups) of parents ...\n ")
        noOfParents2FamilyData = {}
        for nodeID in self.pedigreeGraph:
            parents = self.pedigreeGraph.predecessors(nodeID)
            noOfParents = len(parents)
            if noOfParents not in noOfParents2FamilyData:
                noOfParents2FamilyData[noOfParents] = PassingData(
                    parentTupleSet=set(), parentIDSet=set(), childIDSet=set(),\
                    individualIDSet=set())
            parents.sort()
            noOfParents2FamilyData[noOfParents].parentTupleSet.add(tuple(parents))
            for parentID in parents:
                noOfParents2FamilyData[noOfParents].parentIDSet.add(parentID)
                noOfParents2FamilyData[noOfParents].individualIDSet.add(parentID)
            noOfParents2FamilyData[noOfParents].childIDSet.add(nodeID)
            noOfParents2FamilyData[noOfParents].individualIDSet.add(nodeID)
        
        noOfNuclearFamilies = noOfParents2FamilyData.get(2, 0)
        
        self._reportFamilyStructure(noOfParents2FamilyData)
        return PassingData(noOfParents2FamilyData=noOfParents2FamilyData)
    
    def getFamilyStructurePlinkWay(self, ):
        """
        Plink ignores individuals that do not have independent entries (only show up as parents of others).
        
        Plink Mendel (error detection) works only on families where both parents have their own independent entries in the file.
            This function calculates the number of them (=len(noOfParents2FamilyData[2].parentTupleSet) ).
        
        #founders = number of lines (entries/individuals) where both parents are 0 (do not think in graph way).
        #non-founders = number of lines (entries/individuals), one or both parents are NOT 0.
            non-founders were classified into two
                #non-founders with 2 parents in N1 nuclear families = both their parents are included in the file with independent entries (AKA genotyped).
                    = len(noOfParents2FamilyData[2].childIDSet)
                    N1 = len(noOfParents2FamilyData[2].parentTupleSet)
                    
                #non-founders without 2 parents in N2 nuclear families = the rest
                    = len(noOfParents2FamilyData[1].childIDSet) + len(noOfParents2FamilyData[0].childIDSet)
                    N2 = len(noOfParents2FamilyData[1].parentTupleSet) + len(noOfParents2FamilyData[0].parentTupleSet)
            
        """
        sys.stderr.write("Getting number of unique parent-set that both parents ")
        pGraph = self.pedigreeGraph
        self._resetInput()
        tfamEntryIndividualIDSet = set()
        founderIndividualIDSet = set()
        nonFounderIndividualIDSet = set()
        for row in self:
            if row.paternalID=='0' and row.maternalID=='0':
                founderIndividualIDSet.add(row.individualID)
            else:
                nonFounderIndividualIDSet.add(row.individualID)
            
            tfamEntryIndividualIDSet.add(row.individualID)
        self._resetInput()
        
        noOfParents2FamilyData = {}
        for nodeID in self.pedigreeGraph:
            if nodeID in tfamEntryIndividualIDSet and nodeID in nonFounderIndividualIDSet:
                #must have an independent entry
                #and exclude founders
                parents = self.pedigreeGraph.predecessors(nodeID)
                parents.sort()
                #calculate no of parents in the plink way, both parents must have independent entries
                noOfParents = 0
                for parentID in parents:
                    if parentID in tfamEntryIndividualIDSet:
                        noOfParents += 1
                if noOfParents not in noOfParents2FamilyData:
                    noOfParents2FamilyData[noOfParents] = PassingData(
                        parentTupleSet=set(), parentIDSet=set(), childIDSet=set(), \
                        individualIDSet=set())
                
                parentTupleList = []
                for parentID in parents:
                    if parentID!='0':
                        parentTupleList.append(parentID)
                        noOfParents2FamilyData[noOfParents].parentIDSet.add(parentID)
                        noOfParents2FamilyData[noOfParents].individualIDSet.add(parentID)
                
                noOfParents2FamilyData[noOfParents].parentTupleSet.add(tuple(parentTupleList))
                noOfParents2FamilyData[noOfParents].childIDSet.add(nodeID)
                noOfParents2FamilyData[noOfParents].individualIDSet.add(nodeID)
        
        self._reportFamilyStructure(noOfParents2FamilyData)
        return PassingData(noOfParents2FamilyData=noOfParents2FamilyData,
            founderIndividualIDSet=founderIndividualIDSet,\
            nonFounderIndividualIDSet=nonFounderIndividualIDSet)
    
    def _reportFamilyStructure(self, noOfParents2FamilyData=None):
        """
        2013.07.19
        """
        sys.stderr.write("\t%s\t%s\t%s\t%s\t%s\n"%("parentSetSize",
            "noOfFamilies", "noOfParents", "noOfKids", "noOfUniqueIndividuals"))
        for noOfParents, familyData in noOfParents2FamilyData.items():
            parentIDSet = familyData.parentIDSet
            childIDSet = familyData.childIDSet
            individualIDSet = familyData.individualIDSet
            sys.stderr.write("\t%s\t%s\t%s\t%s\t%s\n"%(
                noOfParents, len(familyData.parentTupleSet), len(parentIDSet), \
                len(childIDSet), len(individualIDSet)))
        
    
    def next(self):
        try:
            row = next(self.csvFile)
        except:
            raise StopIteration
        if not self.isRealCSV:
            row = row.strip().split()
        familyID, individualID, paternalID, maternalID, sex, phenotype = row
        return PassingData(familyID=familyID, individualID=individualID,
            paternalID=paternalID, \
            maternalID=maternalID, sex=sex, phenotype=phenotype)

if __name__ == '__main__':
    main_class = PlinkPedigreeFile
    po = ProcessOptions(sys.argv, main_class.option_default_dict, error_doc=main_class.__doc__)
    instance = main_class(**po.long_option2value)
    instance.run()


#!/usr/bin/env python3
"""
Examples:
    #setup database in postgresql
    %s -v postgresql -u yh -z pdc -d pmdb -k sunset
    
    #setup database in mysql
    %s -u yh -z papaya.usc.edu
    
Description:
    2017.03.17
    This is the ORM architecture of SunsetDB.
"""
import sys, os, math, copy
__doc__ = __doc__%(sys.argv[0], sys.argv[0])

from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, Integer, String, DateTime, func, Float, Text
from sqlalchemy import create_engine
from sqlalchemy import ForeignKey
from sqlalchemy.orm import relationship
from sqlalchemy import UniqueConstraint, create_engine
from sqlalchemy import and_, or_, not_
from sqlalchemy.engine.url import URL
from sqlalchemy.orm import scoped_session, sessionmaker
from sqlalchemy.schema import ThreadLocalMetaData, MetaData
from sqlalchemy.types import LargeBinary

from datetime import datetime

from palos import ProcessOptions, PassingData
from palos.db import Database, TableClass


Base = declarative_base()

_schemaname_ = "sunset"

class Country(Base, TableClass):
    """
    2017-4-17
    """
    __tablename__ = 'country'
    __table_args__ = {'schema':_schemaname_}

    id = Column(Integer,primary_key=True)
    name = Column(String(100), unique=True)
    abbr = Column(String(10))
    capital = Column(Text)
    latitude = Column(Float)
    longitude = Column(Float)
    created_by = Column(String(128))
    updated_by = Column(String(128))
    date_created = Column(DateTime, default=datetime.now())
    date_updated = Column(DateTime)

    site_list = relationship("Site",back_populates="country",cascade='all,delete')
    
class Site(Base, TableClass):
    """
    20170411 add study_id
    2012.12.6 add unique constraint on (latitude, longitude)
    2011-4-29
        add column altitude
    2011-3-1
    2017-4-17
        use new sqlalchemy syntax
    """
    __tablename__ = 'site'
    __table_args__ = {'schema':_schemaname_}

    id = Column(Integer,primary_key=True)
    short_name = Column(String(256))
    description = Column(Text)
    latitude = Column(Float)
    longitude = Column(Float)
    altitude = Column(Float)    #2011-4-29
    study_id = Column(Integer,ForeignKey(_schemaname_+".study.id"))  #2017-04-17
    #study = ManyToOne('%s.Study'%(__name__), colname='study_id', ondelete='CASCADE', onupdate='CASCADE')    #2017.04.11
    city = Column(String(100))
    stateprovince = Column(String(100))
    region = Column(String(100))
    zippostal = Column(String(20))
    country_id = Column(Integer,ForeignKey(_schemaname_+".country.id"))  #2017-04-17
    #country = ManyToOne("%s.Country"%(__name__), colname='country_id', ondelete='CASCADE', onupdate='CASCADE')
    created_by = Column(String(128))
    updated_by = Column(String(128))
    date_created = Column(DateTime, default=datetime.now())
    date_updated = Column(DateTime)
    UniqueConstraint('short_name', 'latitude', 'longitude', 'city', 'stateprovince', 'country_id',name='site_sllcsc')
    UniqueConstraint('short_name', 'latitude', 'longitude',name='site_short_name_latitude_longtude')
    
    study = relationship("Study",back_populates="site_list")
    country = relationship("Country",back_populates="site_list")

class Study(Base,TableClass):
    """
    2013.3.12 table used to group individuals
        table Individual and Ind2Ind refers to this table.
    2017.4.17
        use new sqlalchemy syntax
    """
    __tablename__ = 'study'
    __table_args__ = {'schema':_schemaname_}
    
    id = Column(Integer,primary_key=True)
    short_name = Column(String(256), unique=True)
    description = Column(Text)
    created_by = Column(String(128))
    updated_by = Column(String(128))
    date_created = Column(DateTime, default=datetime.now())
    date_updated = Column(DateTime)
    
    site_list = relationship("Site",back_populates="study",cascade='all,delete')
    

class README(Base, TableClass):
    __tablename__ = 'readme'
    __table_args__ = {'schema':_schemaname_}
    
    id = Column(Integer,primary_key=True)
    
    title = Column(String(2000))
    description = Column(Text)
    created_by = Column(String(128))
    updated_by = Column(String(128))
    date_created = Column(DateTime, default=datetime.now())
    date_updated = Column(DateTime)

class SunsetDB(Database):
    __doc__ = __doc__
    option_default_dict = copy.deepcopy(Database.option_default_dict)
    def __init__(self, **keywords):
        Database.__init__(self, **keywords)
        ##self.setup_engine()    #2012.12.18 it needs __metadata__, __session__ from each db-definition file to be ready. can't be run here.
        self.READMEClass = README    #2012.12.18 required to figure out data_dir
        self.dbID2indEntry = {}

    def getCountry(self, country_name=None, capital=None, abbr=None, latitude=None, longitude=None):
        """
        2011-4-28
        
        """
        db_entry = self.session.query(Country).filter_by(name=country_name).first()
        if not db_entry:
            db_entry = Country(name=country_name, capital=capital, abbr=abbr, latitude=latitude, longitude=longitude)
            self.session.add(db_entry)
            self.session.flush()
        return db_entry
    
    def getSite(self, short_name=None, description=None, city=None, stateprovince=None, region=None, country_name=None, \
            latitude=None, longitude=None, altitude=None, study_id=None):
        """
        20170412 added study_id
        20121206 added argument short_name
        20110428
        """
        if country_name:
            country = self.getCountry(country_name=country_name)
        else:
            country = None
        if short_name :
            short_name = short_name
        elif description:
            short_name = description
        elif city:
            short_name = city
        else:
            short_name = None
        query = self.session.query(Site)
        if country:
            query = query.filter_by(country_id=country.id)
        if short_name:
            query = query.filter_by(short_name=short_name)
        if city:
            query = query.filter_by(city=city)
        if stateprovince:
            query = query.filter_by(stateprovince=stateprovince)
        if region:
            query = query.filter_by(region=region)
        if latitude:
            query = query.filter_by(latitude=latitude)
        if longitude:
            query = query.filter_by(longitude=longitude)
        if altitude:
            query = query.filter_by(altitude=altitude)
        if study_id:
            query = query.filter_by(study_id=study_id)
        db_entry = query.first()
        if not db_entry:
            db_entry = Site(short_name=short_name, description=description, city=city, stateprovince=stateprovince, \
                        region=region, country=country, \
                        latitude=latitude, longitude=longitude, altitude=altitude, study_id=study_id)
            self.session.add(db_entry)
            self.session.flush()
        return db_entry
    
    def getStudy(self, short_name=None, description=None):
        """
        20170412
        """
        db_entry = self.session.query(Study).filter_by(short_name=short_name).first()
        if not db_entry:
            db_entry = Study(short_name=short_name, description=description)
            self.session.add(db_entry)
            self.session.flush()
        return db_entry
    
if __name__ == '__main__':
    main_class = SunsetDB
    po = ProcessOptions(sys.argv, main_class.option_default_dict, error_doc=main_class.__doc__)
    instance = main_class(**po.long_option2value)
    #instance.setup(create_tables=True)
    sys.stderr.write("data_dir is %s.\n"%(instance.data_dir))
    import pdb
    pdb.set_trace()


#!/usr/bin/env python3
"""
A class wrapper for SAM/BAM file. It is an extension of pysam.Samfile.	
    http://wwwfgu.anat.ox.ac.uk/~andreas/documentation/samtools/api.html#pysam.Samfile
"""
import os, sys
from palos.utils import PassingData
import pysam
from pysam import AlignedRead

class BamFile(pysam.Samfile):
    """
    path: the path to a bam file.
    mode: 'rb: bam file. r: sam file.
    2011-7-11 A class wrapper for SAM/BAM file. 
        It is an extension of pysam.Samfile:
        http://wwwfgu.anat.ox.ac.uk/~andreas/documentation/samtools/api.html#pysam.Samfile
    """
    def __init__(self, path, mode, **keywords):
        """
        2011-7-11
        """
        pysam.Samfile.__init__(self, path, mode, **keywords)
        self.path = path
        self.mode = mode
    
    def traverseBamByRead(self, processor=None):
        """
        2011-7-10
            add samfile to param_obj
        2011-2-8
            a traverser used by other functions
        """
        self.seek(0)
        it = self.fetch()
        counter = 0
        real_counter = 0
        qname2count = {}
        param_obj = PassingData(real_counter=real_counter, counter=counter, qname2count=qname2count, samfile=self)
        for read in it:
            counter += 1
            exitCode = processor.run(read, param_obj=param_obj)
                
            if counter%10000==0:
                sys.stderr.write("%s\t%s\t\t%s"%('\x08'*80, param_obj.counter, param_obj.real_counter))
            if exitCode:	#2011-7-8
                break
        processor.qname2count = param_obj.qname2count	#2011-2-9 pass it to the processor
        max_redundant_read_count = max(param_obj.qname2count.values())
        sys.stderr.write("\n %s unique reads among %s mapped reads, max redundant read count=%s. Done.\n"%\
                        (len(param_obj.qname2count), param_obj.real_counter, max_redundant_read_count))

class YHAlignedRead(object):
    def __init__(self, alignedRead=None):
        self.read = alignedRead
        #AlignedRead(self, *keywordList, **keywords)
        
        cigar_code2no_of_bases = {}
        for cigar_tuple in self.read.cigar: #presented as a list of tuples (operation,length).
            #For example, the tuple [ (0,3), (1,5), (0,2) ] 
            #   refers to an alignment with 3 matches, 5 insertions and another 2 matches.
            # CIGAR operation MIDNSHP=X =>012345678
            cigar_code = cigar_tuple[0]
            if cigar_code not in cigar_code2no_of_bases:
                cigar_code2no_of_bases[cigar_code] = 0
            cigar_code2no_of_bases[cigar_code] += cigar_tuple[1]
        self.no_of_matches = cigar_code2no_of_bases.get(0, 0)	#0 represents match.
        self.match_fraction = self.no_of_matches/float(self.read.alen)
        self.no_of_aligned_bases = self.read.alen
        self.no_of_insertions = cigar_code2no_of_bases.get(1, 0)
        self.no_of_deletions = cigar_code2no_of_bases.get(2, 0)
        self.no_of_mismatches = self.read.alen - self.no_of_matches
        
        if 8 in cigar_code2no_of_bases:
            if cigar_code2no_of_bases[8]!=self.no_of_mismatches:
                sys.stderr.write("Error: No of mismatches in cigar (%s) doesn't match the number (%s) calculated.\n"%\
                    (cigar_code2no_of_bases[8], self.no_of_mismatches))
                import pdb
                pdb.set_trace()
        self.cigar_code2no_of_bases = cigar_code2no_of_bases
    
    def getMatchFraction(self):
        """
        2012.10.14
        """
        return self.match_fraction
    
    def getNoOfMatches(self):
        """
        2012.10.14
        """
        return self.no_of_matches
    
    def getNoOfMismatches(self):
        """
        2012.10.14
        """
        return self.no_of_mismatches
    
    def getNoOfInsertions(self):
        """
        2012.10.14
        """
        return self.no_of_insertions
    
    def getNoOfDeletions(self):
        """
        2012.10.14
        """
        return self.no_of_deletions
    
    def getNoOfIndels(self):
        """
        2012.10.14
        """
        return self.no_of_deletions + self.no_of_insertions